#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import math
import warnings
from abc import abstractmethod
from collections import OrderedDict
from collections.abc import Sequence
from itertools import product
from typing import Any, Callable
from unittest import mock, TestCase
from warnings import warn

import torch
from botorch.acquisition.objective import PosteriorTransform
from botorch.exceptions.warnings import (
    BotorchTensorDimensionWarning,
    InputDataWarning,
    NumericsWarning,
)
from botorch.models.model import FantasizeMixin, Model
from botorch.posteriors.gpytorch import GPyTorchPosterior
from botorch.posteriors.posterior import Posterior
from botorch.sampling.base import MCSampler
from botorch.sampling.get_sampler import GetSampler
from botorch.sampling.stochastic_samplers import StochasticSampler
from botorch.test_functions.base import (
    BaseTestProblem,
    ConstrainedBaseTestProblem,
    CorruptedTestProblem,
)
from botorch.test_functions.synthetic import Rosenbrock
from botorch.utils.transforms import unnormalize
from gpytorch.distributions import MultitaskMultivariateNormal, MultivariateNormal
from linear_operator.operators import AddedDiagLinearOperator, DiagLinearOperator
from torch import Tensor


EMPTY_SIZE = torch.Size()


def skip_if_import_error(func: Callable) -> Callable:
    def f(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ImportError as e:
            warn(
                "Skipping test because module is not installed. Received the "
                f"following error: {e}"
            )

    return f


def sample_random_feasible(
    f: BaseTestProblem, dtype: torch.dtype, device: torch.device
) -> Tensor:
    r"""Sample random feasible point for the given test function.

    Args:
        f: The test function instance.
        dtype: The dtype of the random point.
        device: The device of the random point.

    Returns:
        A random feasible point of shape `1 x f.dim`.
    """
    round_ids = f.discrete_inds + f.categorical_inds
    if isinstance(f, ConstrainedBaseTestProblem):
        # Sample a bunch of points and hope that one of them is feasible.
        # We could repeat this in a loop but it is not worth risking the
        # tests hanging forever. If no feasible point is found, we can bypass the test.
        X = unnormalize(
            torch.rand(2**12, f.dim, dtype=dtype, device=device),
            bounds=f.bounds,
        )
        X[..., round_ids] = X[..., round_ids].round()
        feasible = (f.evaluate_slack(X) >= 0).all(dim=-1)
        if feasible.any():
            return X[feasible][0]
        else:  # pragma: no cover
            raise RuntimeError(
                f"No feasible point found for {f.__class__.__name__}. Skipping test."
            )
    X = unnormalize(
        torch.rand(1, f.dim, dtype=dtype, device=device),
        bounds=f.bounds,
    )
    X[..., round_ids] = X[..., round_ids].round()
    return X


class BotorchTestCase(TestCase):
    r"""Basic test case for Botorch.

    This
        1. sets the default device to be `torch.device("cpu")`
        2. ensures that no warnings are suppressed by default.
    """

    device = torch.device("cpu")

    def setUp(self, suppress_input_warnings: bool = True) -> None:
        """Set up the test case.

        Args:
            suppress_input_warnings: If True, suppress common input warnings
                (see below).
        """
        warnings.resetwarnings()
        warnings.simplefilter("always", append=True)
        if suppress_input_warnings:
            warnings.filterwarnings(
                "ignore",
                message="The model inputs are of type",
                category=InputDataWarning,
            )
            warnings.filterwarnings(
                "ignore",
                message="Non-strict enforcement of botorch tensor conventions.",
                category=BotorchTensorDimensionWarning,
            )
            warnings.filterwarnings(
                "ignore",
                message=r"Data \(outcome observations\) is not standardized ",
                category=InputDataWarning,
            )
            warnings.filterwarnings(
                "ignore",
                message=r"Data \(input features\) is not",
                category=InputDataWarning,
            )
            warnings.filterwarnings(
                "ignore",
                message="has known numerical issues",
                category=NumericsWarning,
            )
            warnings.filterwarnings(
                "ignore",
                message="Model converter code is deprecated",
                category=DeprecationWarning,
            )

    def assertAllClose(
        self,
        input: Any,
        other: Any,
        rtol: float = 1e-05,
        atol: float = 1e-08,
        equal_nan: bool = False,
    ) -> None:
        r"""Assert that two tensors are close.

        Calls torch.testing.assert_close, using the signature and default behavior
        of torch.allclose.

        The formula asserted is abs(input - other) <= atol + rtol * abs(other).

        Args:
            input: First tensor or tensor-or-scalar-like to compare
            other: Second tensor or tensor-or-scalar-like to compare
            rtol: Relative tolerance
            atol: Absolute tolerance
            equal_nan: If True, consider NaN values as equal

        Example output:
            AssertionError: Scalars are not close!

            Absolute difference: 1.0000034868717194 (up to 0.0001 allowed)
            Relative difference: 0.8348668001940709 (up to 1e-05 allowed)
        """
        # Why not just use the signature and behavior of `torch.testing.assert_close`?
        # Because we used `torch.allclose` for testing in the past, and the two don't
        # behave exactly the same. In particular, `assert_close` requires both `atol`
        # and `rtol` to be set if either one is.
        torch.testing.assert_close(
            input,
            other,
            rtol=rtol,
            atol=atol,
            equal_nan=equal_nan,
        )


class BaseTestProblemTestCaseMixIn:
    r"""Mixin for testing BaseTestProblem (functions) implementations."""

    def test_forward_and_evaluate_true(self):
        r"""Run every BaseTestProblem in `self.functions` on random inputs.
        Runs both `forward` and `evaluate_true`.
        """
        dtypes = (torch.float, torch.double)
        batch_shapes = (torch.Size(), torch.Size([2]), torch.Size([2, 3]))
        for dtype, batch_shape, f in product(dtypes, batch_shapes, self.functions):
            f.to(device=self.device, dtype=dtype)
            X = torch.rand(*batch_shape, f.dim, device=self.device, dtype=dtype)
            X = f.bounds[0] + X * (f.bounds[1] - f.bounds[0])
            for inds in [f.discrete_inds, f.categorical_inds]:
                X[..., inds] = X[..., inds].round()
            res_forward = f(X)
            res_evaluate_true = f.evaluate_true(X)
            # Evaluating outside bounds should raise
            X_out_of_bounds = f.bounds[1:, :] + 1
            with self.assertRaisesRegex(
                ValueError, "Expected `X` to be within the bounds of the test problem."
            ):
                f(X_out_of_bounds)
            for method, res in {
                "forward": res_forward,
                "evaluate_true": res_evaluate_true,
            }.items():
                with self.subTest(
                    f"{dtype}_{batch_shape}_{f.__class__.__name__}_{method}"
                ):
                    self.assertEqual(res.dtype, dtype)
                    self.assertEqual(res.device.type, self.device.type)
                    tail_shape = torch.Size(
                        [f.num_objectives] if f.num_objectives > 1 else []
                    )
                    self.assertEqual(res.shape, batch_shape + tail_shape)

    @property
    @abstractmethod
    def functions(self) -> Sequence[BaseTestProblem]:
        r"""The functions that should be tested.

        Typically defined as a class attribute on the test case subclassing this class.
        """


class SyntheticTestFunctionTestCaseMixin:
    r"""Mixin for testing synthetic `BaseTestProblem` aka test functions."""

    def test_optimal_value(self):
        """Test that a function's optimal_value is correctly computed,
        and defined if it should be.
        """
        for dtype in (torch.float, torch.double):
            for f in self.functions:
                f.to(device=self.device, dtype=dtype)
                if f._optimal_value is None:
                    with self.assertRaisesRegex(NotImplementedError, "optimal value"):
                        f.optimal_value
                else:
                    optval = f.optimal_value
                    optval_exp = -f._optimal_value if f.negate else f._optimal_value
                    self.assertEqual(optval, optval_exp)

    def test_optimizer(self):
        r"""Test that optimizers are correctly computed and the optimizer value is
        better than the function value at some random point.
        """
        for dtype in (torch.float, torch.double):
            for f in self.functions:
                f.to(device=self.device, dtype=dtype)
                try:
                    Xopt = f.optimizers.clone().requires_grad_(True)
                except NotImplementedError:
                    continue
                res = f(Xopt, noise=False)
                # if we have optimizers, we have the optimal value
                res_exp = torch.full_like(res, f.optimal_value)
                self.assertAllClose(res, res_exp, atol=1e-3, rtol=1e-3)
                if f._check_grad_at_opt:
                    grad = torch.autograd.grad([*res], Xopt)[0]
                    self.assertLess(grad.abs().max().item(), 1e-3)
                # Check that the optimizer is better than (or equal to) a random point.
                try:
                    random_point = sample_random_feasible(
                        f=f, dtype=dtype, device=self.device
                    )
                except RuntimeError:  # pragma: no cover
                    # If no feasible point is found, we can skip the test.
                    # Infeasible points can have better than optimal values.
                    continue
                f_random = f(random_point, noise=False).item()
                f_opt = res[0].item()
                if f.is_minimization_problem:
                    self.assertLessEqual(f_opt, f_random)
                else:
                    self.assertGreaterEqual(f_opt, f_random)

    @property
    @abstractmethod
    def functions(self) -> Sequence[BaseTestProblem]:
        """The functions that should be tested.

        Typically defined as a class attribute on the test case subclassing this class.
        """
        pass  # pragma: no cover


class MultiObjectiveTestProblemTestCaseMixin:
    r"""Mixin for testing multi-objective test problems.

    This class provides test cases for attributes,
    maximum hypervolume, and reference points
    of multi-objective test problems.
    """

    def test_attributes(self):
        r"""Test that each function has the required attributes."""
        for f in self.functions:
            self.assertTrue(hasattr(f, "dim"))
            self.assertTrue(hasattr(f, "num_objectives"))
            self.assertEqual(f.bounds.shape, torch.Size([2, f.dim]))

    def test_max_hv(self):
        r"""Test the maximum hypervolume (max_hv) attribute for each function."""
        for dtype in (torch.float, torch.double):
            for f in self.functions:
                f.to(device=self.device, dtype=dtype)
                if f._max_hv is None:
                    with self.assertRaises(NotImplementedError):
                        f.max_hv
                else:
                    self.assertEqual(f.max_hv, f._max_hv)

    def test_ref_point(self):
        """Test the reference point (ref_point) attribute
        for each function (for each dtype).
        """
        for dtype in (torch.float, torch.double):
            for f in self.functions:
                f.to(dtype=dtype, device=self.device)
                self.assertTrue(
                    torch.allclose(
                        f.ref_point,
                        torch.tensor(f._ref_point, dtype=dtype, device=self.device),
                    )
                )

    @property
    @abstractmethod
    def functions(self) -> Sequence[BaseTestProblem]:
        """The functions that should be tested.

        Typically defined as a class attribute on the test case subclassing this class.
        """
        pass  # pragma: no cover


class ConstrainedTestProblemTestCaseMixin:
    """Mixin for testing constrained test problems.

    This class provides test cases for attributes and methods
    of constrained test problems, including testing the number of
    constraints and the evaluation of constraint slack.
    """

    def test_num_constraints(self):
        """Test that each function has the required num_constraints attribute."""
        for f in self.functions:
            self.assertTrue(hasattr(f, "num_constraints"))

    def test_evaluate_slack(self):
        """Test the evaluate_slack method for each function.

        This test verifies that:

        1. The evaluate_slack_true and evaluate_slack methods
            return tensors of the expected shape

        2. The relationship between evaluate_slack and evaluate_slack_true
        is consistent with the constraint_noise_std attribute of the function
        """
        for dtype in (torch.float, torch.double):
            for f in self.functions:
                f.to(device=self.device, dtype=dtype)
                X = unnormalize(
                    torch.rand(1, f.dim, device=self.device, dtype=dtype),
                    bounds=f.bounds,
                )
                slack_true = f.evaluate_slack_true(X)
                # Mock out the random generator to ensure that noise realizations are
                # sizable so we don't run into any floating point comparison issues.
                with mock.patch(
                    "botorch.test_functions.base.torch.randn_like",
                    side_effect=lambda y: y,
                ):
                    slack_observed = f.evaluate_slack(X)

                self.assertEqual(slack_true.shape, torch.Size([1, f.num_constraints]))
                self.assertEqual(
                    slack_observed.shape, torch.Size([1, f.num_constraints])
                )
                is_equal = (slack_observed == slack_true).bool()
                if isinstance(f.constraint_noise_std, float):
                    self.assertEqual(
                        is_equal.all().item(), f.constraint_noise_std == 0.0
                    )
                elif isinstance(f.constraint_noise_std, list):
                    for i, noise_std in enumerate(f.constraint_noise_std):
                        self.assertEqual(
                            is_equal[:, i].item(), noise_std in (0.0, None)
                        )
                else:
                    self.assertTrue(is_equal.all().item())

    @property
    @abstractmethod
    def functions(self) -> Sequence[BaseTestProblem]:
        r"""The functions that should be tested.

        Typically defined as a class attribute on the test case subclassing this class.
        """
        pass  # pragma: no cover


class TestCorruptedProblemsMixin(BotorchTestCase):
    r"""Mixin for testing corrupted test problems.

    This class provides setup and utility functions
    for testing corrupted test problems using a specified outlier generator
    and a Rosenbrock problem.
    """

    def setUp(self, suppress_input_warnings: bool = True) -> None:
        r"""Set up the test case with a dummy outlier generator
        and a Rosenbrock problem.

        Args:
            suppress_input_warnings: If True, suppress common input warnings.
        """
        super().setUp(suppress_input_warnings=suppress_input_warnings)

        def outlier_generator(
            problem: torch.Tensor | BaseTestProblem, X: Any, bounds: Any
        ) -> torch.Tensor:
            r"""Generate outliers for the given problem.

            Args:
                problem: The test problem.
                X: Input tensor.
                bounds: Bounds for the input.

            Returns:
                A tensor of ones with the same shape as the input.
            """
            return torch.ones(X.shape[0])

        self.outlier_generator = outlier_generator

        self.rosenbrock_problem = CorruptedTestProblem(
            base_test_problem=Rosenbrock(),
            outlier_fraction=1.0,
            outlier_generator=outlier_generator,
            seeds=[1, 2],
        )


class MockPosterior(Posterior):
    r"""This class is used to simulate a posterior with specified mean,
    variance, and samples.

    Everything is deterministic in this class.
    """

    def __init__(
        self,
        mean: torch.Tensor | None = None,
        variance: torch.Tensor | None = None,
        samples: torch.Tensor | None = None,
        base_shape: torch.Size | None = None,
        batch_range: tuple[int, int] | None = None,
    ) -> None:
        r"""Initialize the MockPosterior with specified attributes.

        Args:
            mean: The mean of the posterior.
            variance: The variance of the posterior.
            samples: Samples to return from `rsample`,
                unless `base_samples` is provided.
            base_shape: If given, this is returned as `base_sample_shape`,
                and also used as the base of the `_extended_shape`.
            batch_range: If given, this is returned as `batch_range`.
                Defaults to (0, -2).
        """
        self._mean = mean
        self._variance = variance
        self._samples = samples
        self._base_shape = base_shape
        self._batch_range = batch_range or (0, -2)

    @property
    def device(self) -> torch.device:
        r"""Return the device of the posterior."""
        for t in (self._mean, self._variance, self._samples):
            if torch.is_tensor(t):
                return t.device
        return torch.device("cpu")

    @property
    def dtype(self) -> torch.dtype:
        r"""Return the data type of the posterior."""
        for t in (self._mean, self._variance, self._samples):
            if torch.is_tensor(t):
                return t.dtype
        return torch.float32

    @property
    def batch_shape(self) -> torch.Size:
        r"""Return the batch shape of the posterior."""
        for t in (self._mean, self._variance, self._samples):
            if torch.is_tensor(t):
                return t.shape[:-2]
        raise NotImplementedError  # pragma: no cover

    def _extended_shape(
        self,
        sample_shape: torch.Size = torch.Size(),  # noqa: B008
    ) -> torch.Size:
        r"""Return the extended shape of the posterior."""
        return sample_shape + self.base_sample_shape

    @property
    def base_sample_shape(self) -> torch.Size:
        r"""Return the base sample shape of the posterior."""
        if self._base_shape is not None:
            return self._base_shape
        if self._samples is not None:
            return self._samples.shape
        if self._mean is not None:
            return self._mean.shape
        if self._variance is not None:
            return self._variance.shape
        return torch.Size()

    @property
    def batch_range(self) -> tuple[int, int]:
        r"""Return the batch range of the posterior."""
        return self._batch_range

    @property
    def mean(self):
        r"""Return the mean of the posterior."""
        return self._mean

    @property
    def variance(self):
        r"""Return the variance of the posterior."""
        return self._variance

    def rsample(
        self,
        sample_shape: torch.Size | None = None,
    ) -> Tensor:
        """Return mock samples by extending the shape
        of the initially specified samples.

        Args:
            sample_shape: The shape of the samples to generate.

        Returns:
            A tensor of samples with the specified shape.
        """
        if sample_shape is None:
            sample_shape = torch.Size()
        extended_shape = self._extended_shape(sample_shape)
        return self._samples.expand(extended_shape)

    def rsample_from_base_samples(
        self,
        sample_shape: torch.Size,
        base_samples: Tensor,
    ) -> Tensor:
        if base_samples.shape[: len(sample_shape)] != sample_shape:
            raise RuntimeError(
                "`sample_shape` disagrees with shape of `base_samples`. "
                f"Got {sample_shape=} and {base_samples.shape=}."
            )
        return self.rsample(sample_shape)


@GetSampler.register(MockPosterior)
def get_sampler_mock(
    posterior: MockPosterior, sample_shape: torch.Size, **kwargs: Any
) -> MCSampler:
    """Get a `StochasticSampler` with the specified `sample_shape`.

    Args:
        posterior: Used only for dispatching so that `get_sampler`
            works with a `MockPosterior`.
        sample_shape: The shape of the samples to generate.
        kwargs: Passed to `StochasticSampler`

    Returns:
        A `StochasticSampler` for the mock posterior.
    """
    return StochasticSampler(sample_shape=sample_shape, **kwargs)


class MockModel(Model, FantasizeMixin):
    """Mock ``Model`` that implements dummy methods and feeds through specified outputs.

    Its ``posterior`` is a ``MockPosterior``.
    """

    def __init__(self, posterior: MockPosterior) -> None:  # noqa: D107
        r"""Initialize the MockModel with a specified posterior.

        Args:
            posterior: The mock posterior to use for the model.
        """
        super(Model, self).__init__()
        self._posterior = posterior

    def posterior(
        self,
        X: Tensor,
        output_indices: list[int] | None = None,
        posterior_transform: PosteriorTransform | None = None,
        observation_noise: bool | torch.Tensor = False,
    ) -> MockPosterior:
        r"""Return the posterior of the model.

        Args:
            X: Ignored; present for compatibility with super class.
            output_indices: Ignored; present for compatibility with super class.
            posterior_transform: Optional.
            observation_noise: Ignored; present for compatibility with super class.

        Returns:
            The posterior of the model, possibly transformed.
        """
        if posterior_transform is not None:
            return posterior_transform(self._posterior)
        else:
            return self._posterior

    @property
    def num_outputs(self) -> int:
        r"""Return the number of outputs of the model."""
        extended_shape = self._posterior._extended_shape()
        return extended_shape[-1] if len(extended_shape) > 0 else 0

    @property
    def batch_shape(self) -> torch.Size:
        r"""Return the batch shape of the model."""
        extended_shape = self._posterior._extended_shape()
        return extended_shape[:-2]

    def state_dict(self, *args, **kwargs) -> None:
        """Dummy method, has no effect"""
        pass

    def load_state_dict(
        self, state_dict: OrderedDict | None = None, strict: bool = False
    ) -> None:
        """Dummy method, has no effect.

        Args:
            state_dict: The state dictionary to load.
            strict: Whether to strictly enforce that the keys in state_dict match
                the keys returned by this module's state_dict function.
        """
        pass


class MockAcquisitionFunction:
    r"""Mock acquisition function object that implements dummy methods."""

    def __init__(self):  # noqa: D107
        """
        Initialize the MockAcquisitionFunction.
        This function does not really do anything,
        but it takes an input of shape (b,q,d)
        and returns a tensor of shape (b,).
        """
        self.model = None
        self.X_pending = None
        self._call_args = {"__call__": [], "set_X_pending": []}

    def __call__(self, X):
        self._call_args["__call__"].append(X)
        return X[..., 0].max(dim=-1).values

    def set_X_pending(self, X_pending: Tensor | None = None):
        self._call_args["set_X_pending"].append(X_pending)
        self.X_pending = X_pending


def get_random_data(
    batch_shape: torch.Size, m: int, d: int = 1, n: int = 10, **tkwargs
) -> tuple[Tensor, Tensor]:
    r"""Generate random data for testing purposes.

    Args:
        batch_shape: The batch shape of the data.
        m: The number of outputs.
        d: The dimension of the input.
        n: The number of data points.
        tkwargs: `device` and `dtype` tensor constructor kwargs.

    Returns:
        A tuple `(train_X, train_Y)` with randomly generated training data.
    """
    rep_shape = batch_shape + torch.Size([1, 1])
    train_x = torch.stack(
        [torch.linspace(0, 0.95, n, **tkwargs) for _ in range(d)], dim=-1
    )
    train_x = train_x + 0.05 * torch.rand_like(train_x).repeat(rep_shape)
    train_x[0] += 0.02  # modify the first batch
    train_y = torch.sin(train_x[..., :1] * (2 * math.pi))
    train_y = train_y + 0.2 * torch.randn(n, m, **tkwargs).repeat(rep_shape)
    return train_x, train_y


def get_test_posterior(
    batch_shape: torch.Size,
    q: int = 1,
    m: int = 1,
    interleaved: bool = True,
    lazy: bool = False,
    independent: bool = False,
    **tkwargs,
) -> GPyTorchPosterior:
    r"""Generate a Posterior for testing purposes.

    Args:
        batch_shape: The batch shape of the data.
        q: The number of candidates
        m: The number of outputs.
        interleaved: A boolean indicating the format of the
            MultitaskMultivariateNormal
        lazy: A boolean indicating if the posterior should be lazy
        independent: A boolean indicating whether the outputs are independent
        tkwargs: `device` and `dtype` tensor constructor kwargs.


    """
    if independent:
        mvns = []
        for _ in range(m):
            mean = torch.rand(*batch_shape, q, **tkwargs)
            a = torch.rand(*batch_shape, q, q, **tkwargs)
            covar = a @ a.transpose(-1, -2)
            flat_diag = torch.rand(*batch_shape, q, **tkwargs)
            covar = covar + torch.diag_embed(flat_diag)
            mvns.append(MultivariateNormal(mean, covar))
        mtmvn = MultitaskMultivariateNormal.from_independent_mvns(mvns)
    else:
        mean = torch.rand(*batch_shape, q, m, **tkwargs)
        a = torch.rand(*batch_shape, q * m, q * m, **tkwargs)
        covar = a @ a.transpose(-1, -2)
        flat_diag = torch.rand(*batch_shape, q * m, **tkwargs)
        if lazy:
            covar = AddedDiagLinearOperator(covar, DiagLinearOperator(flat_diag))
        else:
            covar = covar + torch.diag_embed(flat_diag)
        mtmvn = MultitaskMultivariateNormal(mean, covar, interleaved=interleaved)
    return GPyTorchPosterior(mtmvn)


def get_max_violation_of_bounds(samples: torch.Tensor, bounds: torch.Tensor) -> float:
    """
    The maximum value by which samples lie outside bounds.

    A negative value indicates that all samples lie within bounds.

    Args:
        samples: An `n x q x d` - dimension tensor, as might be returned from
            `sample_q_batches_from_polytope`.
        bounds: A `2 x d` tensor of lower and upper bounds for each column.
    """
    n, q, d = samples.shape
    samples = samples.reshape((n * q, d))
    lower = samples.min(0).values
    upper = samples.max(0).values
    lower_dist = (bounds[0, :] - lower).max().item()
    upper_dist = (upper - bounds[1, :]).max().item()
    return max(lower_dist, upper_dist)


def get_max_violation_of_constraints(
    samples: torch.Tensor,
    constraints: list[tuple[Tensor, Tensor, float]] | None,
    equality: bool,
) -> float:
    r"""
    Amount by which equality constraints are not obeyed.

    Args:
        samples: An `n x q x d` - dimension tensor, as might be returned from
            `sample_q_batches_from_polytope`.
        constraints: A list of tuples (indices, coefficients, rhs),
            with each tuple encoding an inequality constraint of the form
            `\sum_i (X[indices[i]] * coefficients[i]) = rhs`, or `>=` if
            `equality` is False.
        equality: Whether these are equality constraints (not inequality).
    """
    n, q, d = samples.shape
    max_error = 0
    if constraints is not None:
        for ind, coef, rhs in constraints:
            if ind.ndim == 1:
                constr = samples[:, :, ind] @ coef
            else:
                constr = samples[:, ind[:, 0], ind[:, 1]] @ coef

            if equality:
                error = (constr - rhs).abs().max()
            else:
                error = (rhs - constr).max()
            max_error = max(max_error, error)
    return max_error



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Utilities for multi-objective acquisition functions.
"""

from __future__ import annotations

import warnings
from collections.abc import Callable
from math import ceil
from typing import Any

import torch
from botorch.acquisition import monte_carlo  # noqa F401
from botorch.acquisition.multi_objective.objective import MCMultiOutputObjective
from botorch.acquisition.utils import _prune_inferior_shared_processing
from botorch.exceptions.errors import UnsupportedError
from botorch.exceptions.warnings import BotorchWarning
from botorch.models.deterministic import GenericDeterministicModel
from botorch.models.model import Model
from botorch.sampling.pathwise.posterior_samplers import get_matheron_path_model
from botorch.utils.multi_objective.box_decompositions.box_decomposition import (
    BoxDecomposition,
)
from botorch.utils.multi_objective.box_decompositions.box_decomposition_list import (
    BoxDecompositionList,
)
from botorch.utils.multi_objective.box_decompositions.dominated import (
    DominatedPartitioning,
)
from botorch.utils.multi_objective.pareto import is_non_dominated
from botorch.utils.sampling import draw_sobol_samples
from pyre_extensions import assert_is_instance
from torch import Tensor


def get_default_partitioning_alpha(num_objectives: int) -> float:
    r"""Determines an approximation level based on the number of objectives.

    If `alpha` is 0, FastNondominatedPartitioning should be used. Otherwise,
    an approximate NondominatedPartitioning should be used with approximation
    level `alpha`.

    Args:
        num_objectives: the number of objectives.

    Returns:
        The approximation level `alpha`.
    """
    if num_objectives <= 4:
        return 0.0
    elif num_objectives > 6:
        warnings.warn(
            "EHVI works best for less than 7 objectives.",
            BotorchWarning,
            stacklevel=3,
        )
    return 10 ** (-2 if num_objectives >= 6 else -3)


def prune_inferior_points_multi_objective(
    model: Model,
    X: Tensor,
    ref_point: Tensor,
    objective: MCMultiOutputObjective | None = None,
    constraints: list[Callable[[Tensor], Tensor]] | None = None,
    num_samples: int = 2048,
    max_frac: float = 1.0,
    marginalize_dim: int | None = None,
) -> Tensor:
    r"""Prune points from an input tensor that are unlikely to be pareto optimal.

    Given a model, an objective, and an input tensor `X`, this function returns
    the subset of points in `X` that have some probability of being pareto
    optimal, better than the reference point, and feasible. This function uses
    sampling to estimate the probabilities, the higher the number of points `n`
    in `X` the higher the number of samples `num_samples` should be to obtain
    accurate estimates.

    Args:
        model: A fitted model. Batched models are currently not supported.
        X: An input tensor of shape `n x d`. Batched inputs are currently not
            supported.
        ref_point: The reference point.
        objective: The objective under which to evaluate the posterior.
        constraints: A list of callables, each mapping a Tensor of dimension
            `sample_shape x batch-shape x q x m` to a Tensor of dimension
            `sample_shape x batch-shape x q`, where negative values imply
            feasibility.
        num_samples: The number of samples used to compute empirical
            probabilities of being the best point.
        max_frac: The maximum fraction of points to retain. Must satisfy
            `0 < max_frac <= 1`. Ensures that the number of elements in the
            returned tensor does not exceed `ceil(max_frac * n)`.
        marginalize_dim: A batch dimension that should be marginalized.
            For example, this is useful when using a batched fully Bayesian
            model.

    Returns:
        A `n' x d` with subset of points in `X`, where

            n' = min(N_nz, ceil(max_frac * n))

        with `N_nz` the number of points in `X` that have non-zero (empirical,
        under `num_samples` samples) probability of being pareto optimal.
    """
    max_points, obj_vals, infeas = _prune_inferior_shared_processing(
        model=model,
        X=X,
        is_moo=True,
        objective=objective,
        constraints=constraints,
        num_samples=num_samples,
        max_frac=max_frac,
        marginalize_dim=marginalize_dim,
    )
    if infeas.any():
        obj_vals[infeas] = ref_point
    pareto_mask = is_non_dominated(obj_vals, deduplicate=False) & (
        obj_vals > ref_point
    ).all(dim=-1)
    probs = pareto_mask.to(dtype=X.dtype).mean(dim=0)
    idcs = probs.nonzero().view(-1)
    if idcs.shape[0] > max_points:
        counts, order_idcs = torch.sort(probs, stable=True, descending=True)
        idcs = order_idcs[:max_points]
    effective_n_w = obj_vals.shape[-2] // X.shape[-2]
    idcs = (idcs / effective_n_w).long().unique()
    return X[idcs]


def compute_sample_box_decomposition(
    pareto_fronts: Tensor,
    partitioning: type[BoxDecomposition] = DominatedPartitioning,
    maximize: bool = True,
    num_constraints: int = 0,
) -> Tensor:
    r"""Computes the box decomposition associated with some sampled optimal
    objectives. This also supports the single-objective and constrained optimization
    setting. An objective `y` is feasible if `y <= 0`.

    To take advantage of batch computations, we pad the hypercell bounds with a
    `2 x (M + K)`-dim Tensor of zeros `[0, 0]`.

    Args:
        pareto_fronts: A `num_pareto_samples x num_pareto_points x M` dim Tensor
            containing the sampled optimal set of objectives.
        partitioning: A `BoxDecomposition` module that is used to obtain the
            hyper-rectangle bounds for integration. In the unconstrained case, this
            gives the partition of the dominated space. In the constrained case, this
            gives the partition of the feasible dominated space union the infeasible
            space.
        maximize: If true, the box-decomposition is computed assuming maximization.
        num_constraints: The number of constraints `K`.

    Returns:
        A `num_pareto_samples x 2 x J x (M + K)`-dim Tensor containing the bounds for
        the hyper-rectangles. The number `J` is the smallest number of boxes needed
        to partition all the Pareto samples.
    """
    tkwargs: dict[str, Any] = {
        "dtype": pareto_fronts.dtype,
        "device": pareto_fronts.device,
    }
    # We will later compute `norm.log_prob(NEG_INF)`, this is `-inf` if `NEG_INF` is
    # too small.
    NEG_INF = -1e10

    if pareto_fronts.ndim != 3:
        raise UnsupportedError(
            "Currently this only supports Pareto fronts of the shape "
            "`num_pareto_samples x num_pareto_points x num_objectives`."
        )

    num_pareto_samples = pareto_fronts.shape[0]
    M = pareto_fronts.shape[-1]
    K = num_constraints
    ref_point = torch.ones(M, **tkwargs) * NEG_INF
    weight = 1.0 if maximize else -1.0

    if M == 1:
        # Only consider a Pareto front with one element.
        extreme_values = assert_is_instance(
            weight * torch.max(weight * pareto_fronts, dim=-2).values, Tensor
        )
        ref_point = weight * ref_point.expand(extreme_values.shape)

        if maximize:
            hypercell_bounds = torch.stack(
                [ref_point, extreme_values], dim=-2
            ).unsqueeze(-1)
        else:
            hypercell_bounds = torch.stack(
                [extreme_values, ref_point], dim=-2
            ).unsqueeze(-1)
    else:
        bd_list = []
        for i in range(num_pareto_samples):
            bd_list = bd_list + [
                partitioning(ref_point=ref_point, Y=weight * pareto_fronts[i, :, :])
            ]

        # `num_pareto_samples x 2 x J x (M + K)`
        hypercell_bounds = (
            BoxDecompositionList(*bd_list).get_hypercell_bounds().movedim(0, 1)
        )

        # If minimizing, then the bounds should be negated and flipped
        if not maximize:
            hypercell_bounds = weight * torch.flip(hypercell_bounds, dims=[1])

    # Add an extra box for the inequality constraint.
    if K > 0:
        # `num_pareto_samples x 2 x (J - 1) x K`
        feasible_boxes = torch.zeros(hypercell_bounds.shape[:-1] + (K,), **tkwargs)

        feasible_boxes[..., 0, :, :] = NEG_INF
        # `num_pareto_samples x 2 x (J - 1) x (M + K)`
        hypercell_bounds = torch.cat([hypercell_bounds, feasible_boxes], dim=-1)

        # `num_pareto_samples x 2 x 1 x (M + K)`
        infeasible_box = torch.zeros(
            hypercell_bounds.shape[:-2] + (1, M + K), **tkwargs
        )
        infeasible_box[..., 1, :, M:] = -NEG_INF
        infeasible_box[..., 0, :, 0:M] = NEG_INF
        infeasible_box[..., 1, :, 0:M] = -NEG_INF

        # `num_pareto_samples x 2 x J x (M + K)`
        hypercell_bounds = torch.cat([hypercell_bounds, infeasible_box], dim=-2)

    # `num_pareto_samples x 2 x J x (M + K)`
    return hypercell_bounds


def random_search_optimizer(
    model: GenericDeterministicModel,
    bounds: Tensor,
    num_points: int,
    maximize: bool,
    pop_size: int = 1024,
    max_tries: int = 10,
) -> tuple[Tensor, Tensor]:
    r"""Optimize a function via random search.

    Args:
        model: The model.
        bounds: A `2 x d`-dim Tensor containing the input bounds.
        num_points: The number of optimal points to be outputted.
        maximize: If true, we consider a maximization problem.
        pop_size: The number of function evaluations per try.
        max_tries: The maximum number of tries.

    Returns:
        A two-element tuple containing

        - A `num_points x d`-dim Tensor containing the collection of optimal inputs.
        - A `num_points x M`-dim Tensor containing the collection of optimal
            objectives.
    """
    tkwargs: dict[str, Any] = {"dtype": bounds.dtype, "device": bounds.device}
    weight = 1.0 if maximize else -1.0
    optimal_inputs = torch.tensor([], **tkwargs)
    optimal_outputs = torch.tensor([], **tkwargs)
    num_tries = 0
    num_found = 0
    ratio = 2
    while ratio > 1 and num_tries < max_tries:
        X = draw_sobol_samples(bounds=bounds, n=pop_size, q=1).squeeze(-2)
        Y = model.posterior(X).mean
        X_aug = torch.cat([optimal_inputs, X], dim=0)
        Y_aug = torch.cat([optimal_outputs, Y], dim=0)
        pareto_mask = is_non_dominated(weight * Y_aug)
        optimal_inputs = X_aug[pareto_mask]
        optimal_outputs = Y_aug[pareto_mask]
        num_found = len(optimal_inputs)
        ratio = ceil(num_points / num_found)
        num_tries = num_tries + 1
    # If maximum number of retries exceeded throw out a runtime error.
    if ratio > 1:
        error_text = f"Only found {num_found} optimal points instead of {num_points}."
        raise RuntimeError(error_text)
    else:
        return optimal_inputs[:num_points], optimal_outputs[:num_points]


def sample_optimal_points(
    model: Model,
    bounds: Tensor,
    num_samples: int,
    num_points: int,
    optimizer: Callable[
        [GenericDeterministicModel, Tensor, int, bool, Any], tuple[Tensor, Tensor]
    ] = random_search_optimizer,
    maximize: bool = True,
    optimizer_kwargs: dict[str, Any] | None = None,
) -> tuple[Tensor, Tensor]:
    r"""Compute a collection of optimal inputs and outputs from samples of a Gaussian
    Process (GP).

    Steps:
    (1) The samples are generated using random Fourier features (RFFs).
    (2) The samples are optimized sequentially using an optimizer.

    TODO: We can generalize the GP sampling step to accommodate for other sampling
        strategies rather than restricting to RFFs e.g. decoupled sampling.

    TODO: Currently this defaults to random search optimization, might want to
        explore some other alternatives.

    Args:
        model: The model. This does not support models which include fantasy
            observations.
        bounds: A `2 x d`-dim Tensor containing the input bounds.
        num_samples: The number of GP samples.
        num_points: The number of optimal points to be outputted.
        optimizer: A callable that solves the deterministic optimization problem.
        maximize: If true, we consider a maximization problem.
        optimizer_kwargs: The additional arguments for the optimizer.

    Returns:
        A two-element tuple containing

        - A `num_samples x num_points x d`-dim Tensor containing the collection of
            optimal inputs.
        - A `num_samples x num_points x M`-dim Tensor containing the collection of
            optimal objectives.
    """
    tkwargs: dict[str, Any] = {"dtype": bounds.dtype, "device": bounds.device}
    M = model.num_outputs
    d = bounds.shape[-1]
    if M == 1:
        if num_points > 1:
            raise UnsupportedError(
                "For single-objective optimization `num_points` should be 1."
            )
    if optimizer_kwargs is None:
        optimizer_kwargs = {}
    pareto_sets = torch.zeros((num_samples, num_points, d), **tkwargs)
    pareto_fronts = torch.zeros((num_samples, num_points, M), **tkwargs)
    for i in range(num_samples):
        sample_i = get_matheron_path_model(model=model)
        ps_i, pf_i = optimizer(
            model=sample_i,
            bounds=bounds,
            num_points=num_points,
            maximize=maximize,
            **optimizer_kwargs,
        )
        pareto_sets[i, ...] = ps_i
        pareto_fronts[i, ...] = pf_i

    return pareto_sets, pareto_fronts



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from botorch.models.transforms.factory import get_rounding_input_transform
from botorch.models.transforms.input import (
    ChainedInputTransform,
    Normalize,
    Round,
    Warp,
)
from botorch.models.transforms.outcome import (
    Bilog,
    ChainedOutcomeTransform,
    Log,
    Power,
    Standardize,
)


__all__ = [
    "get_rounding_input_transform",
    "Bilog",
    "ChainedInputTransform",
    "ChainedOutcomeTransform",
    "Log",
    "Normalize",
    "Power",
    "Round",
    "Standardize",
    "Warp",
]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Helpers for handling input or outcome constraints.
"""

from __future__ import annotations

import math

from collections.abc import Callable

from functools import partial

import torch
from gpytorch import settings
from gpytorch.constraints import Interval

from torch import Tensor


def get_outcome_constraint_transforms(
    outcome_constraints: tuple[Tensor, Tensor] | None,
) -> list[Callable[[Tensor], Tensor]] | None:
    r"""Create outcome constraint callables from outcome constraint tensors.

    Args:
        outcome_constraints: A tuple of `(A, b)`. For `k` outcome constraints
            and `m` outputs at `f(x)``, `A` is `k x m` and `b` is `k x 1` such
            that `A f(x) <= b`.

    Returns:
        A list of callables, each mapping a Tensor of size `b x q x m` to a
        tensor of size `b x q`, where `m` is the number of outputs of the model.
        Negative values imply feasibility. The callables support broadcasting
        (e.g. for calling on a tensor of shape `mc_samples x b x q x m`).

    Example:
        >>> # constrain `f(x)[0] <= 0`
        >>> A = torch.tensor([[1., 0.]])
        >>> b = torch.tensor([[0.]])
        >>> outcome_constraints = get_outcome_constraint_transforms((A, b))
    """
    if outcome_constraints is None:
        return None
    A, b = outcome_constraints

    def _oc(a: Tensor, rhs: Tensor, Y: Tensor) -> Tensor:
        r"""Evaluate constraints.

        Note: einsum multiples Y by a and sums over the `m`-dimension. Einsum
            is ~2x faster than using `(Y * a.view(1, 1, -1)).sum(dim-1)`.

        Args:
            a: `m`-dim tensor of weights for the outcomes
            rhs: Singleton tensor containing the outcome constraint value
            Y: `... x b x q x m` tensor of function values

        Returns:
            A `... x b x q`-dim tensor where negative values imply feasibility
        """
        lhs = torch.einsum("...m, m", [Y, a])
        return lhs - rhs

    return [partial(_oc, a, rhs) for a, rhs in zip(A, b)]


def get_monotonicity_constraints(
    d: int,
    descending: bool = False,
    dtype: torch.dtype | None = None,
    device: torch.device | None = None,
) -> tuple[Tensor, Tensor]:
    """Returns a system of linear inequalities `(A, b)` that generically encodes order
    constraints on the elements of a `d`-dimsensional space, i.e. `A @ x < b` implies
    `x[i] < x[i + 1]` for a `d`-dimensional vector `x`.

    Idea: Could encode `A` as sparse matrix, if it is supported well.

    Args:
        d: Dimensionality of the constraint space, i.e. number of monotonic parameters.
        descending: If True, forces the elements of a vector to be monotonically de-
            creasing and be monotonically increasing otherwise.
        dtype: The dtype of the returned Tensors.
        device: The device of the returned Tensors.

    Returns:
        A tuple of Tensors `(A, b)` representing the monotonicity constraint as a system
        of linear inequalities `A @ x < b`. `A` is `(d - 1) x d`-dimensional and `b` is
        `(d - 1) x 1`-dimensional.
    """
    A = torch.zeros(d - 1, d, dtype=dtype, device=device)
    idx = torch.arange(d - 1)
    A[idx, idx] = 1
    A[idx, idx + 1] = -1
    b = torch.zeros(d - 1, 1, dtype=dtype, device=device)
    if descending:
        A = -A
    return A, b


class NonTransformedInterval(Interval):
    """Modification of the GPyTorch interval class that does not apply transformations.

    This is generally useful, and it is a requirement for the sparse parameters of the
    Relevance Pursuit model [Ament2024pursuit]_, since it is not possible to achieve
    exact zeros with the sigmoid transformations that are applied by default in the
    GPyTorch Interval class. The variant implemented here does not apply transformations
    to the parameters, instead passing the bounds constraint to the scipy L-BFGS
    optimizer. This allows for the expression of exact zeros for sparse optimization
    algorithms.

    NOTE: On a high level, the cleanest solution for this would be to separate out the
    1) definition and book-keeping of parameter constraints on the one hand, and
    2) the re-parameterization of the variables with some monotonic transformation,
    since the two steps are orthogonal, but this would require refactoring GPyTorch.
    """

    def __init__(
        self,
        lower_bound: float | Tensor,
        upper_bound: float | Tensor,
        initial_value: float | Tensor | None = None,
    ):
        """Constructor of the NonTransformedInterval class.

        Args:
            lower_bound: The lower bound of the interval.
            upper_bound: The upper bound of the interval.
            initial_value: The initial value of the parameter.
        """
        super().__init__(
            lower_bound=lower_bound,
            upper_bound=upper_bound,
            transform=None,
            inv_transform=None,
            initial_value=initial_value,
        )

    def transform(self, tensor: Tensor) -> Tensor:
        return tensor

    def inverse_transform(self, transformed_tensor: Tensor) -> Tensor:
        return transformed_tensor


class LogTransformedInterval(Interval):
    """Modification of the GPyTorch interval class.

    The Interval class in GPyTorch will map the parameter to the range [0, 1] before
    applying the inverse transform. LogTransformedInterval skips this step to avoid
    numerical issues, and applies the log transform directly to the parameter values.
    GPyTorch automatically recognizes that the bound constraint have not been applied
    yet, and passes the bounds to the optimizer instead, which then optimizes
    log(parameter) under the constraints log(lower) <= log(parameter) <= log(upper).
    """

    def __init__(
        self,
        lower_bound: float,
        upper_bound: float,
        initial_value: float | None = None,
    ):
        """Constructor of the LogTransformedInterval class.

        Args:
            lower_bound: The lower bound of the interval.
            upper_bound: The upper bound of the interval.
            initial_value: The initial value of the parameter.
        """
        super().__init__(
            lower_bound=lower_bound,
            upper_bound=upper_bound,
            transform=torch.exp,
            inv_transform=torch.log,
            initial_value=initial_value,
        )

        # Save the untransformed initial value
        self.register_buffer(
            "initial_value_untransformed",
            (
                torch.tensor(initial_value).to(self.lower_bound)
                if initial_value is not None
                else None
            ),
        )

        if settings.debug.on():
            max_bound = torch.max(self.upper_bound)
            min_bound = torch.min(self.lower_bound)
            if max_bound == math.inf or min_bound == -math.inf:
                raise RuntimeError(
                    "Cannot make an Interval directly with non-finite bounds. Use a "
                    "derived class like GreaterThan or LessThan instead."
                )

    def transform(self, tensor: Tensor) -> Tensor:
        """Transform the parameter using the exponential function.

        Args:
            tensor: Tensor of parameter values to transform.

        Returns:
            Tensor of transformed parameter values.
        """
        return self._transform(tensor)

    def inverse_transform(self, transformed_tensor: Tensor) -> Tensor:
        """Untransform the parameter using the natural logarithm.

        Args:
            tensor: Tensor of parameter values to untransform.

        Returns:
            Tensor of untransformed parameter values.
        """
        return self._inv_transform(transformed_tensor)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import torch
from botorch.models.gpytorch import GPyTorchModel
from botorch.posteriors.gpytorch import GPyTorchPosterior
from gpytorch.distributions import MultivariateNormal
from linear_operator.operators import IdentityLinearOperator, ZeroLinearOperator
from torch import Tensor


r"""
References

.. [wilson2020sampling]
    J. Wilson, V. Borovitskiy, A. Terenin, P. Mostowsky, and M. Deisenroth. Efficiently
    sampling functions from Gaussian process posteriors. International Conference on
    Machine Learning (2020).

.. [wilson2021pathwise]
    J. Wilson, V. Borovitskiy, A. Terenin, P. Mostowsky, and M. Deisenroth. Pathwise
    Conditioning of Gaussian Processes. Journal of Machine Learning Research (2021).
"""


class LatentKroneckerGPPosterior(GPyTorchPosterior):
    r"""
    Dummy posterior class for a LatentKroneckerGP model.
    Internally calls model._rsample_from_base_samples to draw posterior samples via
    pathwise conditioning aka Matheron's rule [wilson2020sampling, wilson2021pathwise].

    This is necessary because BoTorch instantiates the posterior object before creating
    base samples, whereas pathwise conditioning requires the base samples first to
    calculate the posterior samples. To cache expensive computations, which only have
    to be performed once for the same base samples, the results are stored in the model
    instead of the posterior object, because a new posterior object is created in each
    acquisition function call.
    """

    def __init__(
        self,
        model: GPyTorchModel,
        X: Tensor,
    ) -> None:
        r"""A dummy posterior for LatentKroneckerGP models.

        Args:
            model: The LatentKroneckerGP model to which this posterior belongs to.
            X: A `(batch_shape) x q x d`-dim Tensor, where `d` is the dimension
                of the feature space and `q` is the number of points considered
                jointly, on which the posterior shall be evaluated.
        """
        self._dtype = X.dtype
        self._device = X.device
        self.batch_shape = model.batch_shape
        self.output_batch_shape = torch.broadcast_shapes(
            model.batch_shape, X.shape[:-2]
        )
        self.q = X.shape[-2]
        output_dim = self.q * model.T.shape[-1]
        mean = ZeroLinearOperator(
            *self.output_batch_shape, output_dim, dtype=X.dtype, device=X.device
        )
        covar = IdentityLinearOperator(
            output_dim,
            batch_shape=self.output_batch_shape,
            dtype=X.dtype,
            device=X.device,
        )
        dummy_mvn = MultivariateNormal(mean=mean, covariance_matrix=covar)
        super().__init__(distribution=dummy_mvn)
        self.model = model
        self.X = X
        self._is_mt = True

    @property
    def base_sample_shape(self):
        r"""The shape of a base sample used for constructing posterior samples.

        Overwrites the standard `base_sample_shape` call to inform samplers that
        `n_train_full + n_train + n_test` samples are needed rather than n samples.
        """
        n_train_full = self.model.train_inputs[0].shape[-2] * self.model.T.shape[-1]
        n_train = self.model.train_targets.shape[-1]
        n_test = self.q * self.model.T.shape[-1]
        return self.batch_shape + torch.Size([n_train_full + n_train + n_test])

    @property
    def batch_range(self) -> tuple[int, int]:
        r"""The t-batch range.

        This is used in samplers to identify the t-batch component of the
        `base_sample_shape`. The base samples are expanded over the t-batches to
        provide consistency in the acquisition values, i.e., to ensure that a
        candidate produces same value regardless of its position on the t-batch.
        """
        return (0, -1)

    def _extended_shape(
        self,
        sample_shape: torch.Size = torch.Size(),  # noqa: B008
    ) -> torch.Size:
        r"""Returns the shape of the samples produced by the distribution with
        the given `sample_shape`.
        """
        time_shape = torch.Size([self.model.T.shape[-1]])
        q_shape = torch.Size([self.q])
        return sample_shape + self.output_batch_shape + q_shape + time_shape

    def rsample_from_base_samples(
        self,
        sample_shape: torch.Size,
        base_samples: Tensor,
    ) -> Tensor:
        r"""Sample from the posterior (with gradients) using base samples.

        This is intended to be used with a sampler that produces the corresponding base
        samples, and enables acquisition optimization via Sample Average Approximation.

        Since this posterior is a dummy object, call the model to perform sampling.

        Args:
            sample_shape: A `torch.Size` object specifying the sample shape. To
                draw `n` samples, set to `torch.Size([n])`. To draw `b` batches
                of `n` samples each, set to `torch.Size([b, n])`.
            base_samples: A Tensor of `N(0, I)` base samples of shape
                `sample_shape x base_sample_shape`, typically obtained from
                a `Sampler`. This is used for deterministic optimization.

        Returns:
            Samples from the posterior, a tensor of shape
            `self._extended_shape(sample_shape=sample_shape)`.
        """
        if base_samples.shape[: len(sample_shape)] != sample_shape:
            raise RuntimeError(
                "`sample_shape` disagrees with shape of `base_samples`. "
                f"Got {sample_shape=} and {base_samples.shape=}."
            )

        return self.model._rsample_from_base_samples(self.X, base_samples)

    def rsample(
        self,
        sample_shape: torch.Size | None = None,
    ) -> Tensor:
        r"""Sample from the posterior (with gradients).

        Args:
            sample_shape: A `torch.Size` object specifying the sample shape. To
                draw `n` samples, set to `torch.Size([n])`. To draw `b` batches
                of `n` samples each, set to `torch.Size([b, n])`.

        Returns:
            Samples from the posterior, a tensor of shape
            `self._extended_shape(sample_shape=sample_shape)`.
        """
        if sample_shape is None:
            sample_shape = torch.Size([1])
        base_samples = torch.randn(
            sample_shape + self.base_sample_shape,
            dtype=self.X.dtype,
            device=self.X.device,
        )
        return self.rsample_from_base_samples(sample_shape, base_samples)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Multi-Source Gaussian Process Regression models based on GPyTorch models.

References:

.. [Ca2021ms]
    Candelieri, A., & Archetti, F. (2021).
    Sparsifying to optimize over multiple information sources:
    an augmented Gaussian process based algorithm.
    Structural and Multidisciplinary Optimization, 64, 239-255.

Contributor: andreaponti5
"""

from __future__ import annotations

from copy import deepcopy
from typing import Optional

import torch

from botorch import fit_gpytorch_mll
from botorch.exceptions import InputDataError
from botorch.models import SingleTaskGP
from botorch.models.transforms.input import InputTransform
from botorch.models.transforms.outcome import OutcomeTransform
from botorch.utils import draw_sobol_samples
from gpytorch import ExactMarginalLogLikelihood, Module
from gpytorch.likelihoods.likelihood import Likelihood
from gpytorch.means.mean import Mean
from gpytorch.models import ExactGP
from torch import Tensor


def get_random_x_for_agp(
    n: int,
    bounds: Tensor,
    q: int,
    seed: Optional[int] = None,
):
    r"""Draw qMC samples from the box defined by bounds.
    The function assures that at least one point belong to
    the highest fidelity source (the source associated to bounds[1, -1]).

    Args:
        n: The number of samples.
        bounds: A `2 x d` dimensional tensor specifying box constraints on a
            `d`-dimensional space, where bounds[0, :] and bounds[1, :] correspond
            to lower and upper bounds, respectively. The last dimension represents
            number of sources.
        q: The size of each q-batch.
        seed: The seed used for initializing Owen scrambling. If None (default),
            use a random seed.

    Returns:
        A `n x q x d`-dim tensor of qMC samples from the box
        defined by bounds.
    """
    if n < 1:
        raise InputDataError(
            f"The number of points n should be greater than 0 (given n={n})."
        )
    train_x = draw_sobol_samples(bounds=bounds, n=n, q=q, seed=seed).squeeze(1)
    train_x[..., -1] = torch.round(train_x[..., -1], decimals=0)
    if bounds[1, -1] not in train_x[..., -1]:
        true_idxs = torch.randint(0, n, [max(1, int(n * 0.2))])
        train_x[true_idxs, -1] = bounds[1, -1]
    return train_x


class SingleTaskAugmentedGP(SingleTaskGP):
    r"""A single-task multi-source GP model.

    The Augmented Gaussian Process (AGP) is described in [Ca2021ms]_.
    The basic idea is to select a subset of function evaluations (namely,
    augmenting observations), among those performed so far over all the
    cheap sources, to augment the set of observations from the ground
    truth source. The resulting augmented set of observations is used to
    fit the AGP.
    The set of "augmenting observations" is selected depending on a
    discrepancy measure between GP's predictive mean and GP’s predictive
    uncertainty.
    """

    def __init__(
        self,
        train_X: Tensor,
        train_Y: Tensor,
        train_Yvar: Optional[Tensor] = None,
        m: int = 1,
        likelihood: Optional[Likelihood] = None,
        covar_module: Optional[Module] = None,
        mean_module: Optional[Mean] = None,
        outcome_transform: Optional[OutcomeTransform] = None,
        input_transform: Optional[InputTransform] = None,
    ) -> None:
        r"""
        Args:
            train_X: A `batch_shape x n x (d + 1)` tensor of training features,
                where the additional dimension is for the source parameter.
            train_Y: A `batch_shape x n x m` tensor of training observations.
            train_Yvar: A `batch_shape x n x m` tensor of observed measurement
                noise.
            m: The multiplication factor of the model standard deviation used to select
                points from other sources to add to the Augmented GP.
            likelihood: A likelihood. If omitted, use a standard
                GaussianLikelihood with inferred noise level.
            covar_module: The module computing the covariance (Kernel) matrix.
                If omitted, use a `MaternKernel`.
            mean_module: The mean function to be used. If omitted, use a
                `ConstantMean`.
            outcome_transform: An outcome transform that is applied to the
                training data during instantiation and to the posterior during
                inference (that is, the `Posterior` obtained by calling
                `.posterior` on the model will be on the original scale).
            input_transform: An input transform that is applied in the model's
                forward pass.
        """
        if m <= 0:
            raise InputDataError(f"The value of m must be greater than 0, given m={m}.")
        # Divide train_X and train_Y based on the source
        train_S = train_X[..., -1]
        sources = torch.unique(train_S).int()
        if sources.shape[0] == 1:
            raise InputDataError("AGP is meant to be used with more than one source.")
        train_X = [train_X[torch.where(train_S == s)] for s in sources]
        train_Y = [train_Y[torch.where(train_S == s)] for s in sources]
        if train_Yvar is not None:
            train_Yvar = [train_Yvar[torch.where(train_S == s)] for s in sources]
        self.n_true_points = len(train_X[-1])
        self.max_n_cheap_points = max([len(points) for points in train_X[:-1]])

        # Init and fit a SingleTaskGP for each source
        self.models = [
            self._init_fit_gp(
                train_X[s][..., :-1],
                train_Y[s],
                None if train_Yvar is None else train_Yvar[s],
                likelihood,
                covar_module,
                mean_module,
                outcome_transform,
                input_transform,
            )
            for s in sources
        ]

        # Create the training set for the AGP selecting all
        # the observations from the high fidelity source
        # and the reliable observations from the other sources
        reliable_idxs = [
            _get_reliable_observations(
                self.models[-1], self.models[s], train_X[s][..., :-1], m
            )
            for s in sources[:-1]
        ]
        train_X = torch.cat(
            [
                train_X[s] if s == len(sources) - 1 else train_X[s][reliable_idxs[s]]
                for s in sources
            ]
        )[:, :-1]
        train_Y = torch.cat(
            [
                train_Y[s] if s == len(sources) - 1 else train_Y[s][reliable_idxs[s]]
                for s in sources
            ]
        )
        if train_Yvar is not None:
            train_Yvar = torch.cat(
                [
                    (
                        train_Yvar[s]
                        if s == len(sources) - 1
                        else train_Yvar[s][reliable_idxs[s]]
                    )
                    for s in sources
                ]
            )

        super().__init__(
            train_X,
            train_Y,
            train_Yvar,
            likelihood,
            covar_module,
            mean_module,
            outcome_transform,
            input_transform,
        )

    def _init_fit_gp(
        self,
        train_X: Tensor,
        train_Y: Tensor,
        train_Yvar: Optional[Tensor] = None,
        likelihood: Optional[Likelihood] = None,
        covar_module: Optional[Module] = None,
        mean_module: Optional[Mean] = None,
        outcome_transform: Optional[OutcomeTransform] = None,
        input_transform: Optional[InputTransform] = None,
    ) -> SingleTaskGP:
        r"""Initialize and fit a Single Task GP model.

        Args:
            train_X: A `batch_shape x n x d` tensor of training features.
            train_Y: A `batch_shape x n x m` tensor of training observations.
            train_Y: A `batch_shape x n x m` tensor of training observations.
            likelihood: A likelihood. If omitted, use a standard
                GaussianLikelihood with inferred noise level.
            covar_module: The module computing the covariance (Kernel) matrix.
                If omitted, use a `MaternKernel`.
            mean_module: The mean function to be used. If omitted, use a
                `ConstantMean`.
            outcome_transform: An outcome transform that is applied to the
                training data during instantiation and to the posterior during
                inference (that is, the `Posterior` obtained by calling
                `.posterior` on the model will be on the original scale).
            input_transform: An input transform that is applied in the model's
                forward pass.

        Returns:
            The fitted Single Task GP and its Marginal Log Likelihood.
        """
        gp = SingleTaskGP(
            train_X,
            train_Y,
            train_Yvar,
            likelihood=deepcopy(likelihood),
            covar_module=deepcopy(covar_module),
            mean_module=deepcopy(mean_module),
            outcome_transform=outcome_transform,
            input_transform=input_transform,
        )
        mll = ExactMarginalLogLikelihood(gp.likelihood, gp)
        fit_gpytorch_mll(mll)
        return gp


def _get_reliable_observations(
    trusty_model: ExactGP,
    other_model: ExactGP,
    x: Tensor,
    m: int = 1,
) -> Tensor:
    r"""Get the points whose posterior mean computed with other_model
    is inside m * trusty_model standard deviation.

    Args:
        trusty_model: The GP model of the trust source.
        other_model: The GP model of a lower fidelity source.
        x: A `batch_shape x n x d` tensor of training features.
        m: The moltiplicator factor of the model standard deviation used to select
            points from other sources to add to the Augmented GP.
    Returns:
        A `batch_shape x N x d` tensor of reliable points
    """
    m0_posterior = trusty_model.posterior(x)
    m0_mu = torch.flatten(m0_posterior.mean)
    m0_sigma = torch.sqrt(torch.flatten(m0_posterior.variance))
    m1_posterior = other_model.posterior(x)
    m1_mu = torch.flatten(m1_posterior.mean)

    return torch.where(torch.abs(m0_mu - m1_mu) < m * m0_sigma)[0]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Multi-output extensions of the risk measures, implemented as Monte-Carlo
objectives. Except for MVaR, the risk measures are computed over each
output dimension independently. In contrast, MVaR is computed using the
joint distribution of the outputs, and provides more accurate risk estimates.

References

.. [Prekopa2012MVaR]
    A. Prekopa. Multivariate value at risk and related topics.
    Annals of Operations Research, 2012.

.. [Cousin2013MVaR]
    A. Cousin and E. Di Bernardino. On multivariate extensions of Value-at-Risk.
    Journal of Multivariate Analysis, 2013.

.. [Daulton2022MARS]
    S. Daulton, S, Cakmak, M. Balandat, M. Osborne, E. Zhou, and E. Bakshy.
    Robust multi-objective Bayesian optimization under input noise.
    Proceedings of the 39th International Conference on Machine Learning, 2022.
"""

import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable
from math import ceil

import torch
from botorch.acquisition.multi_objective.objective import (
    IdentityMCMultiOutputObjective,
    MCMultiOutputObjective,
)
from botorch.acquisition.risk_measures import CVaR, RiskMeasureMCObjective, VaR
from botorch.exceptions.errors import UnsupportedError
from botorch.exceptions.warnings import BotorchWarning
from botorch.models.model import Model
from botorch.utils.multi_objective.pareto import is_non_dominated
from botorch.utils.transforms import normalize
from torch import Tensor


class MultiOutputRiskMeasureMCObjective(
    RiskMeasureMCObjective, MCMultiOutputObjective, ABC
):
    r"""Objective transforming the multi-output posterior samples to samples
    of a multi-output risk measure.

    The risk measure is calculated over joint q-batch samples from the posterior.
    If the q-batch includes samples corresponding to multiple inputs, it is assumed
    that first `n_w` samples correspond to first input, second `n_w` samples
    correspond to second input, etc.
    """

    def __init__(
        self,
        n_w: int,
        preprocessing_function: Callable[[Tensor], Tensor] | None = None,
    ) -> None:
        r"""Transform the posterior samples to samples of a risk measure.

        Args:
            n_w: The size of the `w_set` to calculate the risk measure over.
            preprocessing_function: A preprocessing function to apply to the
                samples before computing the risk measure. This can be used to
                remove non-objective outcomes or to align all outcomes for
                maximization. For constrained optimization, this should also
                apply feasibility-weighting to samples. Given a `batch x m`-dim
                tensor of samples, this should return a `batch x m'`-dim tensor.
        """
        super().__init__(n_w=n_w, preprocessing_function=preprocessing_function)

    def _prepare_samples(self, samples: Tensor) -> Tensor:
        r"""Prepare samples for risk measure calculations by scaling and
        separating out the q-batch dimension.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.

        Returns:
            A `sample_shape x batch_shape x q x n_w x m'`-dim tensor of
            prepared samples.
        """
        samples = self.preprocessing_function(samples)
        return samples.view(*samples.shape[:-2], -1, self.n_w, samples.shape[-1])

    @abstractmethod
    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        r"""Calculate the risk measure corresponding to the given samples.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.
            X: A `batch_shape x q x d`-dim tensor of inputs. Ignored.

        Returns:
            A `sample_shape x batch_shape x q x m'`-dim tensor of risk measure samples.
        """
        pass  # pragma: no cover


class MultiOutputExpectation(MultiOutputRiskMeasureMCObjective):
    r"""A multi-output MC expectation risk measure.

    For unconstrained problems, we recommend using the `ExpectationPosteriorTransform`
    instead. `ExpectationPosteriorTransform` directly transforms the posterior
    distribution over `q * n_w` to a posterior of `q` expectations, significantly
    reducing the cost of posterior sampling as a result.
    """

    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        r"""Calculate the expectation of the given samples. Expectation is
        calculated over each `n_w` samples in the q-batch dimension.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.
            X: A `batch_shape x q x d`-dim tensor of inputs. Ignored.

        Returns:
            A `sample_shape x batch_shape x q x m'`-dim tensor of expectation samples.
        """
        prepared_samples = self._prepare_samples(samples)
        return prepared_samples.mean(dim=-2)


class IndependentCVaR(CVaR, MultiOutputRiskMeasureMCObjective):
    r"""The multi-output Conditional Value-at-Risk risk measure that operates on
    each output dimension independently. Since this does not consider the joint
    distribution of the outputs (i.e., that the outputs were evaluated on same
    perturbed input and are not independent), the risk estimates provided by
    `IndependentCVaR` in general are more optimistic than the definition of CVaR
    would suggest.

    The Conditional Value-at-Risk measures the expectation of the worst outcomes
    (small rewards or large losses) with a total probability of `1 - alpha`. It
    is commonly defined as the conditional expectation of the reward function,
    with the condition that the reward is smaller than the corresponding
    Value-at-Risk (also defined below).

    NOTE: Due to the use of a discrete `w_set` of samples, the VaR and CVaR
    calculated here are (possibly biased) Monte-Carlo approximations of the
    true risk measures.
    """

    def _get_sorted_prepared_samples(self, samples: Tensor) -> Tensor:
        r"""Get the prepared samples that are sorted over the `n_w` dimension.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.

        Returns:
            A `sample_shape x batch_shape x q x n_w x m'`-dim tensor of sorted samples.
        """
        prepared_samples = self._prepare_samples(samples)
        return prepared_samples.sort(dim=-2, descending=True).values

    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        r"""Calculate the CVaR corresponding to the given samples.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.
            X: A `batch_shape x q x d`-dim tensor of inputs. Ignored.

        Returns:
            A `sample_shape x batch_shape x q x m'`-dim tensor of CVaR samples.
        """
        sorted_samples = self._get_sorted_prepared_samples(samples)
        return sorted_samples[..., self.alpha_idx :, :].mean(dim=-2)


class IndependentVaR(IndependentCVaR):
    r"""The multi-output Value-at-Risk risk measure that operates on each output
    dimension independently. For the same reasons as `IndependentCVaR`, the risk
    estimates provided by this are in general more optimistic than the definition
    of VaR would suggest.

    Value-at-Risk measures the smallest possible reward (or largest possible loss)
    after excluding the worst outcomes with a total probability of `1 - alpha`. It
    is commonly used in financial risk management, and it corresponds to the
    `1 - alpha` quantile of a given random variable.
    """

    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        r"""Calculate the VaR corresponding to the given samples.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.
            X: A `batch_shape x q x d`-dim tensor of inputs. Ignored.

        Returns:
            A `sample_shape x batch_shape x q x m'`-dim tensor of VaR samples.
        """
        sorted_samples = self._get_sorted_prepared_samples(samples)
        return sorted_samples[..., self.alpha_idx, :]


class MultiOutputWorstCase(MultiOutputRiskMeasureMCObjective):
    r"""The multi-output worst-case risk measure."""

    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
        r"""Calculate the worst-case measure corresponding to the given samples.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.
            X: A `batch_shape x q x d`-dim tensor of inputs. Ignored.

        Returns:
            A `sample_shape x batch_shape x q x m'`-dim tensor of worst-case samples.
        """
        prepared_samples = self._prepare_samples(samples)
        return prepared_samples.min(dim=-2).values


class MVaR(MultiOutputRiskMeasureMCObjective):
    r"""The multivariate Value-at-Risk as introduced in [Prekopa2012MVaR]_.

    MVaR is defined as the non-dominated set of points in the extended domain
    of the random variable that have multivariate CDF greater than or equal to
    `alpha`. Note that MVaR is set valued and the size of the set depends on the
    particular realizations of the random variable. [Cousin2013MVaR]_ instead
    propose to use the expectation of the set-valued MVaR as the multivariate
    VaR. We support this alternative with an `expectation` flag.

    This supports approximate gradients as discussed in [Daulton2022MARS]_.
    """

    _verify_output_shape = False

    def __init__(
        self,
        n_w: int,
        alpha: float,
        expectation: bool = False,
        preprocessing_function: Callable[[Tensor], Tensor] | None = None,
        *,
        pad_to_n_w: bool = False,
        filter_dominated: bool = True,
        use_counting: bool = False,
    ) -> None:
        r"""The multivariate Value-at-Risk.

        Args:
            n_w: The size of the `w_set` to calculate the risk measure over.
            alpha: The risk level of MVaR, float in `(0.0, 1.0]`. Each MVaR value
                dominates `alpha` fraction of all observations.
            expectation: If True, returns the expectation of the MVaR set as is
                done in [Cousin2013MVaR]_. Otherwise, it returns the union of all
                values in the MVaR set. Default: False.
            preprocessing_function: A preprocessing function to apply to the
                samples before computing the risk measure. This can be used to
                remove non-objective outcomes or to align all outcomes for
                maximization. For constrained optimization, this should also
                apply feasibility-weighting to samples. Given a `batch x m`-dim
                tensor of samples, this should return a `batch x m'`-dim tensor.
            pad_to_n_w: If True, instead of padding up to `k'`, which is the size of
                the largest MVaR set across all batches, we pad the MVaR set up to
                `n_w`. This produces a return tensor of known size, however, it may
                in general be much larger than the alternative. See `forward` for
                more details on the return shape.
                NOTE: this is only relevant if `expectation=False`.
            filter_dominated: If True, returns the non-dominated subset of
                alpha level points (this is MVaR as defined by [Prekopa2012MVaR]_).
                Disabling this will make it faster, and may be preferable if
                the dominated points will be filtered out later, e.g., while
                calculating the hypervolume. Disabling this is not recommended
                if `expectation=True`.
            use_counting: If True, uses `get_mvar_set_via_counting` for finding the
                MVaR set. This is method is less memory intensive than the vectorized
                implementation, which is beneficial when `n_w` is quite large.
        """
        super().__init__(n_w=n_w, preprocessing_function=preprocessing_function)
        if not 0 < alpha <= 1:
            raise ValueError("`alpha` must be in (0.0, 1.0]")
        self.alpha = alpha
        self.expectation = expectation
        self.pad_to_n_w = pad_to_n_w
        self.filter_dominated = filter_dominated
        self.use_counting = use_counting

    def get_mvar_set_via_counting(self, Y: Tensor) -> list[Tensor]:
        r"""Find MVaR set based on the definition in [Prekopa2012MVaR]_.

        This first calculates the CDF for each point on the extended domain of the
        random variable (the grid defined by the given samples), then takes the
        values with CDF equal to (rounded if necessary) `alpha`. The non-dominated
        subset of these form the MVaR set.

        This implementation processes each batch of `Y` in a for loop using a counting
        based implementation. It requires less memory than the vectorized implementation
        and should be used with large (>128) `n_w` values.

        Args:
            Y: A `batch x n_w x m`-dim tensor of outcomes.

        Returns:
            A `batch` length list of `k x m`-dim tensor of MVaR values, where `k`
            depends on the corresponding batch inputs. Note that MVaR values in general
            are not in-sample points.
        """
        if Y.dim() == 3:
            return sum((self.get_mvar_set_via_counting(y_) for y_ in Y), [])
        m = Y.shape[-1]
        # Generate sets of all unique values in each output dimension.
        # Note that points in MVaR are bounded from above by the
        # independent VaR of each objective. Hence, we only need to
        # consider the unique outcomes that are less than or equal to
        # the VaR of the independent objectives
        var_alpha_idx = ceil(self.alpha * self.n_w) - 1
        Y_sorted = Y.topk(Y.shape[0] - var_alpha_idx, dim=0, largest=False).values
        unique_outcomes_list = []
        for i in range(m):
            sorted_i = Y_sorted[:, i].cpu().clone(memory_format=torch.contiguous_format)
            unique_outcomes_list.append(sorted_i.unique().tolist()[::-1])
        # Convert this into a list of m dictionaries mapping values to indices.
        unique_outcomes = [
            dict(zip(outcomes, range(len(outcomes))))
            for outcomes in unique_outcomes_list
        ]
        # Initialize a tensor counting the number of points in Y that a given grid point
        # is dominated by. This will essentially be a non-normalized CDF.
        counter_tensor = torch.zeros(
            [len(outcomes) for outcomes in unique_outcomes],
            dtype=torch.long,
            device=Y.device,
        )
        # populate the tensor, counting the dominated points.
        # we only need to consider points in Y where at least one
        # objective is less than the max objective value in
        # unique_outcomes_list
        max_vals = torch.tensor(
            [o[0] for o in unique_outcomes_list], dtype=Y.dtype, device=Y.device
        )
        mask = (Y < max_vals).any(dim=-1)
        counter_tensor += self.n_w - mask.sum()
        Y_pruned = Y[mask]
        for y_ in Y_pruned:
            starting_idcs = [unique_outcomes[i].get(y_[i].item(), 0) for i in range(m)]
            slices = tuple(slice(s_idx, None) for s_idx in starting_idcs)
            counter_tensor[slices] += 1

        # Get the count alpha-level points should have.
        alpha_count = ceil(self.alpha * self.n_w)
        # Get the alpha level indices.
        alpha_level_indices = (counter_tensor == alpha_count).nonzero(as_tuple=False)
        # If there are no exact alpha level points, get the smallest alpha' > alpha
        # and find the corresponding alpha level indices.
        if alpha_level_indices.numel() == 0:
            min_greater_than_alpha = counter_tensor[counter_tensor > alpha_count].min()
            alpha_level_indices = (counter_tensor == min_greater_than_alpha).nonzero(
                as_tuple=False
            )
        unique_outcomes = [
            torch.as_tensor(list(outcomes.keys()), device=Y.device, dtype=Y.dtype)
            for outcomes in unique_outcomes
        ]
        alpha_level_points = torch.stack(
            [
                unique_outcomes[i][alpha_level_indices[:, i]]
                for i in range(len(unique_outcomes))
            ],
            dim=-1,
        )
        # MVaR is simply the non-dominated subset of alpha level points.
        if self.filter_dominated:
            mask = is_non_dominated(alpha_level_points)
            mvar = alpha_level_points[mask]
        else:
            mvar = alpha_level_points
        return [mvar]

    def get_mvar_set_vectorized(self, Y: Tensor) -> list[Tensor]:
        r"""Find MVaR set based on the definition in [Prekopa2012MVaR]_.

        This first calculates the CDF for each point on the extended domain of the
        random variable (the grid defined by the given samples), then takes the
        values with CDF equal to (rounded if necessary) `alpha`. The non-dominated
        subset of these form the MVaR set.

        This implementation uses computes the CDF of each point using highly vectorized
        operations. As such, it may use large amounts of memory, particularly when the
        batch size and/or `n_w` are large. It is typically faster than the alternative
        implementation when computing MVaR of a large batch of points with small to
        moderate (<128 for m=2, <64 for m=3) `n_w`.

        Args:
            Y: A `batch x n_w x m`-dim tensor of observations.

        Returns:
            A `batch` length list of `k x m`-dim tensor of MVaR values, where `k`
            depends on the corresponding batch inputs. Note that MVaR values in general
            are not in-sample points.
        """
        if Y.dim() == 2:
            Y = Y.unsqueeze(0)
        batch, m = Y.shape[0], Y.shape[-1]
        # Note that points in MVaR are bounded from above by the
        # independent VaR of each objective. Hence, we only need to
        # consider the unique outcomes that are less than or equal to
        # the VaR of the independent objectives
        var_alpha_idx = ceil(self.alpha * self.n_w) - 1
        n_points = Y.shape[-2] - var_alpha_idx
        Y_sorted = Y.topk(n_points, dim=-2, largest=False).values
        # `y_grid` is the grid formed by all inputs in each batch.
        if m == 2:
            # This is significantly faster but only works with m=2.
            y_grid = torch.stack(
                [
                    Y_sorted[..., 0].repeat_interleave(repeats=n_points, dim=-1),
                    Y_sorted[..., 1].repeat(1, n_points),
                ],
                dim=-1,
            )
        else:
            y_grid = torch.stack(
                [
                    torch.stack(
                        torch.meshgrid(
                            [Y_sorted[b, :, i] for i in range(m)], indexing="ij"
                        ),
                        dim=-1,
                    ).view(-1, m)
                    for b in range(batch)
                ],
                dim=0,
            )
        # Get the non-normalized CDF.
        cdf = (Y.unsqueeze(-2) >= y_grid.unsqueeze(-3)).all(dim=-1).sum(dim=-2)
        # Get the alpha level points
        alpha_count = ceil(self.alpha * self.n_w)
        # NOTE: Need to loop here since mvar may have different shapes.
        mvar = []
        for b in range(batch):
            alpha_level_points = y_grid[b][cdf[b] == alpha_count]
            # If there are no exact alpha level points, get the smallest alpha' > alpha
            # and find the corresponding alpha level indices.
            if alpha_level_points.numel() == 0:
                min_greater_than_alpha = cdf[b][cdf[b] > alpha_count].min()
                alpha_level_points = y_grid[b][cdf[b] == min_greater_than_alpha]
            # MVaR is the non-dominated subset of alpha level points.
            if self.filter_dominated:
                mask = is_non_dominated(alpha_level_points)
                mvar.append(alpha_level_points[mask])
            else:
                mvar.append(alpha_level_points)
        return mvar

    def make_differentiable(self, prepared_samples: Tensor, mvars: Tensor) -> Tensor:
        r"""An experimental approach for obtaining the gradient of the MVaR via
        component-wise mapping to original samples. See [Daulton2022MARS]_.

        Args:
            prepared_samples: A `(sample_shape * batch_shape * q) x n_w x m`-dim tensor
                of posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.
            mvars: A `(sample_shape * batch_shape * q) x k x m`-dim tensor
                of padded MVaR values.
        Returns:
            The same `mvars` with entries mapped to inputs to produce gradients.
        """
        samples = prepared_samples.unsqueeze(-2).repeat(1, 1, mvars.shape[-2], 1)
        mask = samples == mvars.unsqueeze(-3)
        samples[~mask] = 0
        return samples.sum(dim=-3) / mask.sum(dim=-3)

    def forward(
        self,
        samples: Tensor,
        X: Tensor | None = None,
    ) -> Tensor:
        r"""Calculate the MVaR corresponding to the given samples.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.
            X: A `batch_shape x q x d`-dim tensor of inputs. Ignored.

        Returns:
            A `sample_shape x batch_shape x q x m'`-dim tensor of MVaR values,
            if `self.expectation=True`.
            Otherwise, this returns a `sample_shape x batch_shape x (q * k') x m'`-dim
            tensor, where `k'` is the maximum `k` across all batches that is returned
            by `get_mvar_set_...`. Each `(q * k') x m'` corresponds to the `k` MVaR
            values for each `q` batch of `n_w` inputs, padded up to `k'` by repeating
            the last element. If `self.pad_to_n_w`, we set `k' = self.n_w`, producing
            a deterministic return shape.
        """
        batch_shape, m = samples.shape[:-2], samples.shape[-1]
        prepared_samples = self._prepare_samples(samples)
        # This is -1 x n_w x m.
        prepared_samples = prepared_samples.reshape(-1, *prepared_samples.shape[-2:])
        with torch.no_grad():
            if self.use_counting:
                mvar_set = self.get_mvar_set_via_counting(prepared_samples)
            else:
                mvar_set = self.get_mvar_set_vectorized(prepared_samples)
        # Set the `pad_size` to either `self.n_w` or the size of the largest MVaR set.
        pad_size = self.n_w if self.pad_to_n_w else max([_.shape[0] for _ in mvar_set])
        padded_mvar_list = []
        for mvar_ in mvar_set:
            if self.expectation:
                padded_mvar_list.append(mvar_.mean(dim=0))
            else:
                # Repeat the last entry to make `mvar_set` `pad_size x m`.
                repeats_needed = pad_size - mvar_.shape[0]
                padded_mvar_list.append(
                    torch.cat([mvar_, mvar_[-1].expand(repeats_needed, m)], dim=0)
                )
        mvars = torch.stack(padded_mvar_list, dim=0)
        if samples.requires_grad:
            mvars = self.make_differentiable(
                prepared_samples=prepared_samples, mvars=mvars
            )
        return mvars.view(*batch_shape, -1, m)


class MARS(VaR, MultiOutputRiskMeasureMCObjective):
    r"""MVaR Approximation based on Random Scalarizations as introduced
    in [Daulton2022MARS]_.

    This approximates MVaR via VaR of Chebyshev scalarizations, where each
    scalarization corresponds to a point in the MVaR set. As implemented,
    this uses one set of scalarization weights to approximate a single MVaR value.
    Note that due to the normalization within the Chebyshev scalarization,
    the output of this risk measure may not be on the same scale as its inputs.
    """

    _is_mo: bool = False

    def __init__(
        self,
        alpha: float,
        n_w: int,
        chebyshev_weights: Tensor | list[float],
        baseline_Y: Tensor | None = None,
        ref_point: Tensor | list[float] | None = None,
        preprocessing_function: Callable[[Tensor], Tensor] | None = None,
    ) -> None:
        r"""Transform the posterior samples to samples of a risk measure.

        Args:
            alpha: The risk level, float in `(0.0, 1.0]`.
            n_w: The size of the perturbation set to calculate the risk measure over.
            chebyshev_weights: The weights to use in the Chebyshev scalarization.
                The Chebyshev scalarization is applied before computing VaR.
                The weights must be non-negative. See `preprocessing_function` to
                support minimization objectives.
            baseline_Y: An `n' x d`-dim tensor of baseline outcomes to use in
                determining the normalization bounds for Chebyshev scalarization.
                It is recommended to set this via `set_baseline_Y` helper.
            ref_point: An optional MVaR reference point to use in determining
                the normalization bounds for Chebyshev scalarization.
            preprocessing_function: A preprocessing function to apply to the
                samples before computing the risk measure. This can be used to
                remove non-objective outcomes or to align all outcomes for
                maximization. For constrained optimization, this should also
                apply feasibility-weighting to samples.
        """
        if preprocessing_function is None:
            preprocessing_function = IdentityMCMultiOutputObjective()
        super().__init__(
            alpha=alpha,
            n_w=n_w,
            preprocessing_function=preprocessing_function,
        )
        self.chebyshev_weights = torch.as_tensor(chebyshev_weights)
        self.baseline_Y = baseline_Y
        self.register_buffer(
            "ref_point", torch.as_tensor(ref_point) if ref_point is not None else None
        )
        self.mvar = MVaR(n_w=self.n_w, alpha=self.alpha)
        self._chebyshev_objective = None

    def set_baseline_Y(
        self,
        model: Model | None,
        X_baseline: Tensor | None,
        Y_samples: Tensor | None = None,
    ) -> None:
        r"""Set the `baseline_Y` based on the MVaR predictions of the `model`
        for `X_baseline`.

        Args:
            model: The model being used for MARS optimization. Must have a compatible
                `InputPerturbation` transform attached. Ignored if `Y_samples` is given.
            X_baseline: An `n x d`-dim tensor of previously evaluated points.
                Ignored if `Y_samples` is given.
            Y_samples: An optional `(n * n_w) x d`-dim tensor of predictions. If given,
                instead of sampling from the model, these are used.
        """
        if Y_samples is None:
            with torch.no_grad():
                Y = model.posterior(X_baseline.unsqueeze(-2)).mean.squeeze(-2)
        else:
            if model is not None or X_baseline is not None:
                warnings.warn(
                    "`model` and `X_baseline` are ignored when `Y_samples` is "
                    "provided to `MARS.set_baseline_Y`.",
                    BotorchWarning,
                    stacklevel=2,
                )
            Y = Y_samples
        Y = self.preprocessing_function(Y)
        Y = self.mvar(Y).view(-1, Y.shape[-1])
        Y = Y[is_non_dominated(Y)]
        self.baseline_Y = Y

    @property
    def chebyshev_weights(self) -> Tensor:
        r"""The weights used in Chebyshev scalarization."""
        return self._chebyshev_weights

    @chebyshev_weights.setter
    def chebyshev_weights(self, chebyshev_weights: Tensor | list[float]) -> None:
        r"""Update the Chebyshev weights.

        Invalidates the cached Chebyshev objective.

        Args:
            chebyshev_weights: The weights to use in the Chebyshev scalarization.
                The Chebyshev scalarization is applied before computing VaR.
                The weights must be non-negative. See `preprocessing_function` to
                support minimization objectives.
        """
        self._chebyshev_objective = None
        chebyshev_weights = torch.as_tensor(chebyshev_weights)
        if torch.any(chebyshev_weights < 0):
            raise UnsupportedError("Negative weights are not supported in MARS.")
        if chebyshev_weights.dim() != 1:
            raise UnsupportedError("Batched weights are not supported in MARS.")
        self.register_buffer("_chebyshev_weights", chebyshev_weights)

    @property
    def baseline_Y(self) -> Tensor | None:
        r"""Baseline outcomes used in determining the normalization bounds."""
        return self._baseline_Y

    @baseline_Y.setter
    def baseline_Y(self, baseline_Y: Tensor | None) -> None:
        r"""Update the baseline outcomes.

        Invalidates the cached Chebyshev objective.

        Args:
            baseline_Y: An `n' x d`-dim tensor of baseline outcomes to use in
                determining the normalization bounds for Chebyshev scalarization.
                It is recommended to set this via `set_baseline_Y` helper.
        """
        self._chebyshev_objective = None
        self.register_buffer("_baseline_Y", baseline_Y)

    @property
    def chebyshev_objective(self) -> Callable[[Tensor, Tensor | None], Tensor]:
        r"""The objective for applying the Chebyshev scalarization."""
        if self._chebyshev_objective is None:
            self._construct_chebyshev_objective()
        return self._chebyshev_objective

    def _construct_chebyshev_objective(self) -> None:
        r"""Construct a Chebyshev scalarization. Outcomes are first normalized to [0,1],
        then the Chebyshev scalarization is applied.

        NOTE: This is a modified version of the `get_chebyshev_scalarization` helper.
        It doesn't support negative weights. All objectives should be aligned for
        maximization using `preprocessing_function`.
        """
        if self.baseline_Y is None:
            raise RuntimeError(
                "baseline_Y must be set before constructing the Chebyshev objective."
            )
        ref_point = self.ref_point
        if ref_point is not None:
            ref_point = ref_point.to(self.baseline_Y)
        Y_bounds = self._get_Y_normalization_bounds(
            Y=self.baseline_Y, ref_point=ref_point
        )
        if ref_point is not None:
            ref_point = normalize(ref_point.unsqueeze(0), bounds=Y_bounds).squeeze(0)

        def chebyshev_obj(Y: Tensor, X: Tensor | None = None) -> Tensor:
            Y = self.preprocessing_function(Y)
            Y = normalize(Y, bounds=Y_bounds)
            if ref_point is not None:
                Y = Y - ref_point
            product = torch.einsum("...m,m->...m", Y, self.chebyshev_weights.to(Y))
            return product.min(dim=-1).values

        self._chebyshev_objective = chebyshev_obj

    def _prepare_samples(self, samples: Tensor) -> Tensor:
        r"""Prepare samples for VaR computation by applying the Chebyshev scalarization
        and separating out the q-batch dimension.

        Args:
            samples: A `sample_shape x batch_shape x (q * n_w) x m`-dim tensor of
                posterior samples. The q-batches should be ordered so that each
                `n_w` block of samples correspond to the same input.

        Returns:
            A `sample_shape x batch_shape x q x n_w`-dim tensor of prepared samples.
        """
        samples = self.chebyshev_objective(samples)
        return samples.view(*samples.shape[:-1], -1, self.n_w)

    @staticmethod
    def _get_Y_normalization_bounds(
        Y: Tensor,
        ref_point: Tensor | None = None,
    ) -> Tensor:
        r"""Get normalization bounds for scalarizations.

        Args:
            Y: A `n x m`-dim tensor of outcomes.
            ref_point: The reference point.

        Returns:
            A `2 x m`-dim tensor containing the normalization bounds.
        """
        if ref_point is not None:
            ref_point = ref_point.to(Y)

        if Y.ndim != 2:
            raise UnsupportedError("Batched Y is not supported.")

        if Y.shape[-2] == 0:
            # If there are no observations, return standard bounds.
            Y_bounds = torch.zeros(2, Y.shape[-1], dtype=Y.dtype, device=Y.device)
            Y_bounds[1] = 1.0
            return Y_bounds

        pareto_Y = Y[is_non_dominated(Y)]
        if pareto_Y.shape[-2] == 1:
            if ref_point is not None and (pareto_Y > ref_point).all():
                Y_bounds = torch.cat([ref_point.unsqueeze(0), pareto_Y], dim=0)
            else:
                # If there is only one observation, set the bounds to be [Y_m, Y_m + 1]
                # for each objective m. This ensures we do not divide by zero.
                Y_bounds = torch.cat([pareto_Y, pareto_Y + 1], dim=0)
        else:
            if ref_point is None:
                better_than_ref = torch.ones(
                    pareto_Y.shape[0], device=pareto_Y.device, dtype=torch.long
                )
            else:
                better_than_ref = (pareto_Y > ref_point).all(dim=-1)
            if ref_point is not None and better_than_ref.any():
                nadir = ref_point
                pareto_Y = pareto_Y[better_than_ref]
            else:
                nadir = pareto_Y.min(dim=-2).values
            ideal = pareto_Y.max(dim=-2).values
            Y_bounds = torch.stack([nadir, ideal])

        # If any of the lower bounds is equal to the upper bound, increase the
        # upper bound to prevent division by zero.
        Y_range = Y_bounds.max(dim=0).values - Y_bounds.min(dim=0).values
        mask = Y_range <= 0
        Y_bounds[1, mask] = Y_bounds[1, mask] + 1.0
        return Y_bounds



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Functionality for allocating the inducing points of sparse Gaussian
process models.

References

.. [chen2018dpp]
    Laming Chen and Guoxin Zhang and Hanning Zhou, Fast greedy MAP inference
    for determinantal point process to improve recommendation diversity,
    Proceedings of the 32nd International Conference on Neural Information
    Processing Systems, 2018, https://arxiv.org/abs/1709.05135.

"""

from __future__ import annotations

from abc import ABC, abstractmethod

import torch
from botorch.exceptions.errors import UnsupportedError
from botorch.models.model import Model

from botorch.utils.probability.utils import ndtr as Phi, phi
from gpytorch.module import Module
from linear_operator.operators import LinearOperator
from torch import Tensor

NEG_INF = torch.tensor(float("-inf"))


class InducingPointAllocator(ABC):
    r"""
    This class provides functionality to initialize the inducing point locations
    of an inducing point-based model, e.g. a `SingleTaskVariationalGP`.
    """

    @abstractmethod
    def _get_quality_function(
        self,
    ) -> QualityFunction:
        """
        Build the quality function required for this inducing point allocation strategy.

        Returns:
            A quality function.
        """

    def allocate_inducing_points(
        self,
        inputs: Tensor,
        covar_module: Module,
        num_inducing: int,
        input_batch_shape: torch.Size,
    ) -> Tensor:
        r"""
        Initialize the `num_inducing` inducing point locations according to a
        specific initialization strategy. todo say something about quality

        Args:
            inputs: A (\*batch_shape, n, d)-dim input data tensor.
            covar_module: GPyTorch Module returning a LinearOperator kernel matrix.
            num_inducing: The maximun number (m) of inducing points (m <= n).
            input_batch_shape: The non-task-related batch shape.

        Returns:
            A (\*batch_shape, m, d)-dim tensor of inducing point locations.
        """
        quality_function = self._get_quality_function()
        covar_module = covar_module.to(inputs.device)

        # We use 'no_grad' here because `inducing_points` are not
        # auto-differentiable with respect to the kernel hyper-parameters,
        # because `_pivoted_cholesky_init` does in-place operations.
        with torch.no_grad():
            # Evaluate lazily because this may only be needed to figure out what
            # case we are in
            possibly_lazy_kernel = covar_module(inputs)

        base_case = possibly_lazy_kernel.ndimension() == 2
        multi_task_case = (
            possibly_lazy_kernel.ndimension() == 3 and len(input_batch_shape) == 0
        )

        if base_case or multi_task_case:
            train_train_kernel = possibly_lazy_kernel.evaluate_kernel()

        if base_case:
            quality_scores = quality_function(inputs)
            inducing_points = _pivoted_cholesky_init(
                train_inputs=inputs,
                kernel_matrix=train_train_kernel,
                max_length=num_inducing,
                quality_scores=quality_scores,
            )
            return inducing_points

        if multi_task_case:
            input_element = inputs[0] if inputs.ndimension() == 3 else inputs
            kernel_element = train_train_kernel[0]
            quality_scores = quality_function(input_element)
            inducing_points = _pivoted_cholesky_init(
                train_inputs=input_element,
                kernel_matrix=kernel_element,
                max_length=num_inducing,
                quality_scores=quality_scores,
            )
            return inducing_points

        # batched input cases
        batched_inputs = (
            inputs.expand(*input_batch_shape, -1, -1)
            if inputs.ndimension() == 2
            else inputs
        )
        reshaped_inputs = batched_inputs.flatten(end_dim=-3)
        inducing_points = []
        for input_element in reshaped_inputs:
            # the extra kernel evals are a little wasteful but make it
            # easier to infer the task batch size
            # We use 'no_grad' here because `inducing_points` are not
            # auto-differentiable with respect to the kernel hyper-parameters,
            # because `_pivoted_cholesky_init` does in-place operations.
            with torch.no_grad():
                kernel_element = covar_module(input_element).evaluate_kernel()
            # handle extra task batch dimension
            kernel_element = (
                kernel_element[0]
                if kernel_element.ndimension() == 3
                else kernel_element
            )
            quality_scores = quality_function(input_element)
            inducing_points.append(
                _pivoted_cholesky_init(
                    train_inputs=input_element,
                    kernel_matrix=kernel_element,
                    max_length=num_inducing,
                    quality_scores=quality_scores,
                )
            )
        inducing_points = torch.stack(inducing_points).view(
            *input_batch_shape, num_inducing, -1
        )

        return inducing_points


class QualityFunction(ABC):
    """A function that scores inputs with respect
    to a specific criterion."""

    @abstractmethod
    def __call__(self, inputs: Tensor) -> Tensor:  # [n, d] -> [n]
        """
        Args:
            inputs: inputs (of shape n x d)

        Returns:
            A tensor of quality scores for each input, of shape [n]
        """


class UnitQualityFunction(QualityFunction):
    """
    A function returning ones for each element. Using this quality function
    for inducing point allocation corresponds to allocating inducing points
    with the sole aim of minimizing predictive variance, i.e. the approach
    of [burt2020svgp]_.
    """

    @torch.no_grad()
    def __call__(self, inputs: Tensor) -> Tensor:  # [n, d]-> [n]
        """
        Args:
            inputs: inputs (of shape n x d)

        Returns:
            A tensor of ones for each input, of shape [n]
        """
        return torch.ones([inputs.shape[0]], device=inputs.device, dtype=inputs.dtype)


class ExpectedImprovementQualityFunction(QualityFunction):
    """
    A function measuring the quality of input points as their expected
    improvement with respect to a conservative baseline. Expectations
    are according to the model from the previous BO step. See [moss2023ipa]_
    for details and justification.
    """

    def __init__(self, model: Model, maximize: bool):
        r"""
        Args:
            model: The model fitted during the previous BO step. For now, this
                must be a single task model (i.e. num_outputs=1).
            maximize: Set True if we are performing function maximization, else
                set False.
        """
        if model.num_outputs != 1:
            raise NotImplementedError(
                "Multi-output models are currently not supported. "
            )
        self._model = model
        self._maximize = maximize

    @torch.no_grad()
    def __call__(self, inputs: Tensor) -> Tensor:  # [n, d] -> [n]
        """
        Args:
            inputs: inputs (of shape n x d)

        Returns:
            A tensor of quality scores for each input, of shape [n]
        """

        posterior = self._model.posterior(inputs)
        mean = posterior.mean.squeeze(-2).squeeze(-1)  # removing redundant dimensions
        sigma = posterior.variance.clamp_min(1e-12).sqrt().view(mean.shape)

        best_f = torch.max(mean) if self._maximize else torch.min(mean)
        u = (mean - best_f) / sigma if self._maximize else -(mean - best_f) / sigma
        return sigma * (phi(u) + u * Phi(u))


class GreedyVarianceReduction(InducingPointAllocator):
    r"""
    The inducing point allocator proposed by [burt2020svgp]_, that
    greedily chooses inducing point locations with maximal (conditional)
    predictive variance.
    """

    def _get_quality_function(
        self,
    ) -> QualityFunction:
        """
        Build the unit quality function required for the greedy variance
        reduction inducing point allocation strategy.

        Returns:
            A quality function.
        """

        return UnitQualityFunction()


class GreedyImprovementReduction(InducingPointAllocator):
    r"""
    An inducing point allocator that greedily chooses inducing points with large
    predictive variance and that are in promising regions of the search
    space (according to the model form the previous BO step), see [moss2023ipa]_.
    """

    def __init__(self, model: Model, maximize: bool):
        r"""

        Args:
            model: The model fitted during the previous BO step.
            maximize: Set True if we are performing function maximization, else
                set False.
        """
        self._model = model
        self._maximize = maximize

    def _get_quality_function(
        self,
    ) -> QualityFunction:
        """
        Build the improvement-based quality function required for the greedy
        improvement reduction inducing point allocation strategy.

        Returns:
            A quality function.
        """

        return ExpectedImprovementQualityFunction(self._model, self._maximize)


def _pivoted_cholesky_init(
    train_inputs: Tensor,
    kernel_matrix: Tensor | LinearOperator,
    max_length: int,
    quality_scores: Tensor,
    epsilon: float = 1e-6,
) -> Tensor:
    r"""
    A pivoted Cholesky initialization method for the inducing points,
    originally proposed in [burt2020svgp]_ with the algorithm itself coming from
    [chen2018dpp]_. Code is a PyTorch version from [chen2018dpp]_, based on
    https://github.com/laming-chen/fast-map-dpp/blob/master/dpp.py but with a small
    modification to allow the underlying DPP to be defined through its diversity-quality
    decomposition,as discussed by [moss2023ipa]_. This method returns a greedy
    approximation of the MAP estimate of the specified DPP, i.e. its returns a
    set of points that are highly diverse (according to the provided kernel_matrix)
    and have high quality (according to the provided quality_scores).

    Args:
        train_inputs: training inputs (of shape n x d)
        kernel_matrix: kernel matrix on the training inputs
        max_length: number of inducing points to initialize
        quality_scores: scores representing the quality of each candidate
            input (of shape [n])
        epsilon: numerical jitter for stability.

    Returns:
        max_length x d tensor of the training inputs corresponding to the top
        max_length pivots of the training kernel matrix
    """

    # this is numerically equivalent to iteratively performing a pivoted cholesky
    # while storing the diagonal pivots at each iteration
    # TODO: use gpytorch's pivoted cholesky instead once that gets an exposed list
    # TODO: ensure this works in batch mode, which it does not currently.

    # todo test for shape of quality function

    if quality_scores.shape[0] != train_inputs.shape[0]:
        raise ValueError(
            "_pivoted_cholesky_init requires a quality score for each of train_inputs"
        )

    if kernel_matrix.requires_grad:
        raise UnsupportedError(
            "`_pivoted_cholesky_init` does not support using a `kernel_matrix` "
            "with `requires_grad=True`."
        )

    item_size = kernel_matrix.shape[-2]
    cis = torch.zeros(
        (max_length, item_size), device=kernel_matrix.device, dtype=kernel_matrix.dtype
    )
    di2s = kernel_matrix.diagonal()
    scores = di2s * torch.square(quality_scores)
    selected_item = torch.argmax(scores)
    selected_items = [selected_item]

    while len(selected_items) < max_length:
        k = len(selected_items) - 1
        ci_optimal = cis[:k, selected_item]
        di_optimal = torch.sqrt(di2s[selected_item])
        elements = kernel_matrix[..., selected_item, :]
        eis = (elements - torch.matmul(ci_optimal, cis[:k, :])) / di_optimal
        cis[k, :] = eis
        di2s = di2s - eis.pow(2.0)
        di2s[selected_item] = NEG_INF
        scores = di2s * torch.square(quality_scores)
        selected_item = torch.argmax(scores)
        if di2s[selected_item] < epsilon:
            break
        selected_items.append(selected_item)

    ind_points = train_inputs[torch.stack(selected_items)]

    return ind_points[:max_length, :]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from botorch.acquisition.multi_objective.analytic import ExpectedHypervolumeImprovement
from botorch.acquisition.multi_objective.base import (
    MultiObjectiveAnalyticAcquisitionFunction,
    MultiObjectiveMCAcquisitionFunction,
)
from botorch.acquisition.multi_objective.hypervolume_knowledge_gradient import (
    qHypervolumeKnowledgeGradient,
    qMultiFidelityHypervolumeKnowledgeGradient,
)
from botorch.acquisition.multi_objective.logei import (
    qLogExpectedHypervolumeImprovement,
    qLogNoisyExpectedHypervolumeImprovement,
)
from botorch.acquisition.multi_objective.max_value_entropy_search import (
    qMultiObjectiveMaxValueEntropy,
)
from botorch.acquisition.multi_objective.monte_carlo import (
    qExpectedHypervolumeImprovement,
    qNoisyExpectedHypervolumeImprovement,
)
from botorch.acquisition.multi_objective.multi_fidelity import MOMF
from botorch.acquisition.multi_objective.objective import (
    IdentityMCMultiOutputObjective,
    MCMultiOutputObjective,
    WeightedMCMultiOutputObjective,
)
from botorch.acquisition.multi_objective.utils import (
    get_default_partitioning_alpha,
    prune_inferior_points_multi_objective,
)


__all__ = [
    "ExpectedHypervolumeImprovement",
    "get_default_partitioning_alpha",
    "IdentityMCMultiOutputObjective",
    "MCMultiOutputObjective",
    "MOMF",
    "MultiObjectiveAnalyticAcquisitionFunction",
    "MultiObjectiveMCAcquisitionFunction",
    "prune_inferior_points_multi_objective",
    "qExpectedHypervolumeImprovement",
    "qHypervolumeKnowledgeGradient",
    "qLogExpectedHypervolumeImprovement",
    "qLogNoisyExpectedHypervolumeImprovement",
    "qMultiFidelityHypervolumeKnowledgeGradient",
    "qMultiObjectiveMaxValueEntropy",
    "qNoisyExpectedHypervolumeImprovement",
    "WeightedMCMultiOutputObjective",
]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Dummy classes and other helpers that are used in multiple test files
should be defined here to avoid relative imports.
"""

from __future__ import annotations

import math
from typing import Any

import torch
from botorch.acquisition.objective import PosteriorTransform
from botorch.exceptions.errors import UnsupportedError
from botorch.models import SingleTaskGP
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.models.gpytorch import BatchedMultiOutputGPyTorchModel, GPyTorchModel
from botorch.models.model import FantasizeMixin, Model
from botorch.models.model_list_gp_regression import ModelListGP
from botorch.models.transforms.outcome import Standardize
from botorch.models.utils import add_output_dim
from botorch.models.utils.assorted import fantasize
from botorch.posteriors.torch import TorchPosterior
from botorch.utils.datasets import MultiTaskDataset, SupervisedDataset
from gpytorch.distributions.multivariate_normal import MultivariateNormal
from gpytorch.kernels import RBFKernel, ScaleKernel
from gpytorch.likelihoods.gaussian_likelihood import (
    FixedNoiseGaussianLikelihood,
    GaussianLikelihood,
)
from gpytorch.means import ConstantMean
from gpytorch.models.exact_gp import ExactGP
from torch import Size, Tensor
from torch.nn.functional import pad


def get_model(
    train_X: Tensor,
    train_Y: Tensor,
    standardize_model: bool = False,
    use_model_list: bool = False,
) -> SingleTaskGP | ModelListGP:
    num_objectives = train_Y.shape[-1]

    if standardize_model:
        if use_model_list:
            outcome_transform = Standardize(m=1)
        else:
            outcome_transform = Standardize(m=num_objectives)
    else:
        outcome_transform = None

    if use_model_list:
        model = ModelListGP(
            *[
                SingleTaskGP(
                    train_X=train_X,
                    train_Y=train_Y[:, i : i + 1],
                    outcome_transform=outcome_transform,
                )
                for i in range(num_objectives)
            ]
        )
    else:
        model = SingleTaskGP(
            train_X=train_X,
            train_Y=train_Y,
            outcome_transform=outcome_transform,
        )

    return model


def get_fully_bayesian_model(
    train_X: Tensor,
    train_Y: Tensor,
    num_models: int,
    standardize_model: bool = False,
    infer_noise: bool = True,
    **tkwargs: Any,
) -> SaasFullyBayesianSingleTaskGP:
    num_objectives = train_Y.shape[-1]

    if standardize_model:
        outcome_transform = Standardize(m=num_objectives)
    else:
        outcome_transform = None
    mcmc_samples = _get_mcmc_samples(
        num_samples=num_models,
        dim=train_X.shape[-1],
        infer_noise=infer_noise,
        **tkwargs,
    )
    train_Yvar = None if infer_noise else torch.full_like(train_Y, 0.01)

    model = SaasFullyBayesianSingleTaskGP(
        train_X=train_X,
        train_Y=train_Y,
        train_Yvar=train_Yvar,
        outcome_transform=outcome_transform,
    )
    model.load_mcmc_samples(mcmc_samples)

    return model


def _get_mcmc_samples(
    num_samples: int, dim: int, infer_noise: bool, **tkwargs
) -> dict[str, Tensor]:
    mcmc_samples = {
        "lengthscale": 1 + torch.rand(num_samples, 1, dim, **tkwargs),
        "outputscale": 1 + torch.rand(num_samples, **tkwargs),
        "mean": torch.randn(num_samples, **tkwargs),
    }
    if infer_noise:
        mcmc_samples["noise"] = torch.rand(num_samples, 1, **tkwargs)

    return mcmc_samples


def get_fully_bayesian_model_list(
    train_X: Tensor,
    train_Y: Tensor,
    num_models: int,
    standardize_model: bool,
    infer_noise: bool,
    **tkwargs: Any,
) -> ModelListGP:
    model = ModelListGP(
        *[
            get_fully_bayesian_model(
                train_X, train_Y, num_models, standardize_model, infer_noise, **tkwargs
            )
            for _ in range(2)
        ]
    )
    return model


def get_sample_moments(samples: Tensor, sample_shape: Size) -> tuple[Tensor, Tensor]:
    """Computes the mean and covariance of a set of samples.

    Args:
        samples: A tensor of shape `sample_shape x batch_shape x q`.
        sample_shape: The sample_shape input used while generating the samples using
            the pathwise sampling API.
    """
    sample_dim = len(sample_shape)
    samples = samples.view(-1, *samples.shape[sample_dim:])
    loc = samples.mean(dim=0)
    residuals = (samples - loc).permute(*range(1, samples.ndim), 0)
    return loc, (residuals @ residuals.transpose(-2, -1)) / sample_shape.numel()


def standardize_moments(
    transform: Standardize,
    loc: Tensor,
    covariance_matrix: Tensor,
) -> tuple[Tensor, Tensor]:
    """Standardizes the loc and covariance_matrix using the mean and standard
    deviations from a Standardize transform.
    """
    m = transform.means.squeeze().unsqueeze(-1)
    s = transform.stdvs.squeeze().reciprocal().unsqueeze(-1)
    loc = s * (loc - m)
    correlation_matrix = s.unsqueeze(-1) * covariance_matrix * s.unsqueeze(-2)
    return loc, correlation_matrix


def gen_multi_task_dataset(
    yvar: float | None = None,
    task_values: list[int] | None = None,
    skip_task_features_in_datasets: bool = False,
    **tkwargs,
) -> tuple[MultiTaskDataset, tuple[Tensor, Tensor, Tensor | None]]:
    """Constructs a multi-task dataset with two tasks, each with 10 data points.

    Args:
        yvar: The noise level to use for `train_Yvar`. If None, uses `train_Yvar=None`.
        task_values: The values of the task features. If None, uses [0, 1].
        skip_task_features_in_datasets: If True, the task features are not included in
            Xs of the datasets used to construct the datasets. This is useful for
            testing `MultiTaskDataset`.
    """
    if task_values is not None and skip_task_features_in_datasets:
        raise UnsupportedError(  # pragma: no cover
            "`task_values` and `skip_task_features_in_datasets` can't be used together."
        )
    X = torch.linspace(0, 0.95, 10, **tkwargs) + 0.05 * torch.rand(10, **tkwargs)
    X = X.unsqueeze(dim=-1)
    Y1 = torch.sin(X * (2 * math.pi)) + torch.randn_like(X) * 0.2
    Y2 = torch.cos(X * (2 * math.pi)) + torch.randn_like(X) * 0.2
    if task_values is None:
        task_values = [0, 1]
    train_X = torch.cat([pad(X, (1, 0), value=i) for i in task_values])
    train_Y = torch.cat([Y1, Y2])

    Yvar1 = None if yvar is None else torch.full_like(Y1, yvar)
    Yvar2 = None if yvar is None else torch.full_like(Y2, 2 * yvar)
    train_Yvar = None if yvar is None else torch.cat([Yvar1, Yvar2])
    Y3 = torch.tan(X * (2 * math.pi)) + torch.randn_like(X) * 0.2
    Yvar3 = None if yvar is None else torch.full_like(Y3, yvar)
    if len(task_values) == 3:
        train_Y = torch.cat([train_Y, Y3])
        if train_Yvar is not None:
            train_Yvar = torch.cat([train_Yvar, Yvar3])
    feature_slice = slice(1, None) if skip_task_features_in_datasets else slice(None)
    datasets = [
        SupervisedDataset(
            X=train_X[:10, feature_slice],
            Y=Y1,
            Yvar=Yvar1,
            feature_names=["task", "X"][feature_slice],
            outcome_names=["y"],
        ),
        SupervisedDataset(
            X=train_X[10:20, feature_slice],
            Y=Y2,
            Yvar=Yvar2,
            feature_names=["task", "X"][feature_slice],
            outcome_names=["y1"],
        ),
    ]
    if len(task_values) == 3:
        datasets.append(
            SupervisedDataset(
                X=train_X[20:, feature_slice],
                Y=Y3,
                Yvar=Yvar3,
                feature_names=["task", "X"][feature_slice],
                outcome_names=["y2"],
            )
        )
    dataset = MultiTaskDataset(
        datasets=datasets,
        target_outcome_name="y",
        task_feature_index=None if skip_task_features_in_datasets else 0,
    )
    return dataset, (train_X, train_Y, train_Yvar)


def get_pvar_expected(
    posterior: TorchPosterior, model: Model, X: Tensor, m: int
) -> Tensor:
    """Computes the expected variance of a posterior after adding the
    predictive noise from the likelihood.

    Args:
        posterior: The posterior to compute the variance of. Must be a
            `TorchPosterior` object.
        model: The model that generated the posterior. If `m > 1`, this must be
            a `BatchedMultiOutputGPyTorchModel`.
        X: The test inputs.
        m: The number of outputs.

    Returns:
        The expected variance of the posterior after adding the observation
        noise from the likelihood.
    """
    X = model.transform_inputs(X)
    lh_kwargs = {}
    odim = -1  # this is the output dimension index

    if m > 1:
        if not isinstance(model, BatchedMultiOutputGPyTorchModel):
            raise UnsupportedError(
                "`get_pvar_expected` only supports `BatchedMultiOutputGPyTorchModel`s."
            )
        # We need to add a batch dimension to the input to be compatible with the
        # augmented batch shape of the model. This also changes the output dimension
        # index.
        X, odim = add_output_dim(X=X, original_batch_shape=model._input_batch_shape)

    if isinstance(model.likelihood, FixedNoiseGaussianLikelihood):
        noise = model.likelihood.noise.mean(dim=-1, keepdim=True)
        broadcasted_shape = torch.broadcast_shapes(noise.shape, X.shape[:-1])
        lh_kwargs["noise"] = noise.expand(broadcasted_shape)

    pvar_exp = model.likelihood(model(X), X, **lh_kwargs).variance
    if m == 1:
        pvar_exp = pvar_exp.unsqueeze(-1)
    pvar_exp = torch.stack(
        [pvar_exp.select(dim=odim, index=i) for i in range(m)], dim=-1
    )

    # If the model has an outcome transform, we need to untransform the
    # variance according to that transform.
    if hasattr(model, "outcome_transform"):
        _, pvar_exp = model.outcome_transform.untransform(
            Y=torch.zeros_like(pvar_exp), Yvar=pvar_exp, X=X
        )

    return pvar_exp


class DummyNonScalarizingPosteriorTransform(PosteriorTransform):
    scalarize = False

    def evaluate(self, Y):
        pass  # pragma: no cover

    def forward(self, posterior):
        pass  # pragma: no cover


class SimpleGPyTorchModel(GPyTorchModel, ExactGP, FantasizeMixin):
    last_fantasize_flag: bool = False

    def __init__(self, train_X, train_Y, outcome_transform=None, input_transform=None):
        r"""
        Args:
            train_X: A tensor of inputs, passed to self.transform_inputs.
            train_Y: Passed to outcome_transform.
            outcome_transform: Transform applied to train_Y.
            input_transform: A Module that performs the input transformation, passed to
                self.transform_inputs.
        """
        with torch.no_grad():
            transformed_X = self.transform_inputs(
                X=train_X, input_transform=input_transform
            )
        if outcome_transform is not None:
            outcome_transform.train()
            train_Y, _ = outcome_transform(train_Y)
        self._validate_tensor_args(transformed_X, train_Y)
        train_Y = train_Y.squeeze(-1)
        likelihood = GaussianLikelihood()
        super().__init__(train_X, train_Y, likelihood)
        self.mean_module = ConstantMean()
        self.covar_module = ScaleKernel(RBFKernel())
        if outcome_transform is not None:
            self.outcome_transform = outcome_transform
        if input_transform is not None:
            self.input_transform = input_transform
        self._num_outputs = 1
        self.to(train_X)
        self.transformed_call_args = []

    def forward(self, x):
        self.last_fantasize_flag = fantasize.on()
        if self.training:
            x = self.transform_inputs(x)
        self.transformed_call_args.append(x)
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)
